#include "include/block_map.h"
#include "include/spu_utils.h"
#include "include/bl_f_d.h"
#include <string.h>
#include "include/block_lib.h"

// this file is compiled both for floats and doubles

static void zero(BL_DTYPE *x, int n)
{
	memset(x, n*sizeof(BL_DTYPE), 0);
}

static void copy(BL_DTYPE *s, BL_DTYPE *d, int n)
{
	memcpy(d,s,n*sizeof(BL_DTYPE));
}

#ifndef BL_MP_COMB
static PrivateArray *lowMirrorBuf = NULL;
static PrivateArray *highMirrorBuf = NULL;

static void allocPArrs() // allocates parrs ones and then keeps them (the will just take 32kiB + some smaller structs of main memory)
{
	if(!lowMirrorBuf)
	{
		Name name;
		name.procedure = NAME_OFFSET + NAME_OVERLAP;
		name.relative = 0;
		BL_DTYPE* _lowMirrorBuf;
		BL_DTYPE* _highMirrorBuf;

		lowMirrorBuf = new_PrivateArray(name,BL_NSTYPE,BL_CHSIZE,(Addr64)_lowMirrorBuf);
		name.relative++;
		highMirrorBuf = new_PrivateArray(name,BL_NSTYPE,BL_CHSIZE,(Addr64)_highMirrorBuf);
	}
}
#endif

void BL_TN(DistOverlappedMapLocalFunc)(void (*map_func)(BL_DTYPE*, BL_DTYPE*, int, int), BlockDistArray* x, BlockDistArray* res, int N, int overlap, enum bl_edge_value edge)
{
	startTimer(&bl_time.total);
	int i;
	int rank = NestStep_get_rank();
	int groupSize = NestStep_get_size();

	BL_DTYPE data[BL_CHSIZE*3];
	BL_DTYPE *chStart = &data[BL_CHSIZE];
	BL_DTYPE *chOverStart = &data[BL_CHSIZE*2];
	BL_DTYPE lastBuf[BL_CHSIZE];
	
	if(overlap % BL_VSIZE)
		overlap += (BL_VSIZE - overlap % BL_VSIZE);
	if(overlap + BL_VSIZE > N/groupSize ||overlap + BL_VSIZE > BL_CHSIZE)
		printf("Error: Overlap to big for array after adjustment (%d)\n", overlap + BL_VSIZE);

#ifndef BL_MP_COMB
	allocPArrs();
#endif	
	// init BarrF handlers
	// source
	BArr_Handler baX;
	init_BArr_Get(&baX, x, BL_CHSIZE);

	// target
	BArr_Handler baRes;
	init_BArr_Put(&baRes, res, BL_CHSIZE);
	NestStep_step();
	{
#ifdef BL_MP_COMB
		// Blocklib message passing version
		get_sw_BArr(&baX); // get first
		startTimer(&bl_time.comb);
		// prime edges
		// some of the copying and strange offsets in the code below is due to the 
		// fact that dma transfers has to be 16 byte aligned...
		if(groupSize > 1) // send and receive edges
		{
			int size;
			int alignOff;
			int sendUp = (rank != groupSize-1 || edge == BL_CYCLIC);
			int recvFromUp = (rank != groupSize-1 || edge == BL_CYCLIC);
			int sendDown = (rank != 0 || edge == BL_CYCLIC);
			int recvFromDown = (rank != 0 || edge == BL_CYCLIC);
			int upperRank = (rank == groupSize-1 ? 0 : rank +1);
			int lowerRank = (rank == 0 ? groupSize-1 : rank -1);
			if(baX.part%BL_VSIZE)
				alignOff = BL_VSIZE-baX.part%BL_VSIZE;
			else
				alignOff = 0;
			
			if(sendUp) // send upwards
			{
				// Read last chunk from node's part of array. This chunk can have any alignment
				// read some extra to handle that.
				get_BArr_dA(x, &chOverStart[BL_CHSIZE - (overlap+BL_VSIZE)], baX.start + baX.part - (overlap+BL_VSIZE) + alignOff, baX.start + baX.part -1);
				wait_BArr_complete(x, (Addr64)&chOverStart[BL_CHSIZE - (overlap+BL_VSIZE)]);
				// move to fix 16 byte alignment
				copy(chOverStart, &lastBuf[alignOff], BL_CHSIZE);
				// send upwards
				blSendMsg(&lastBuf[BL_CHSIZE - (overlap+BL_VSIZE)], (overlap+BL_VSIZE)*sizeof(BL_DTYPE), upperRank);
			}
			if(recvFromDown)
			{
				blRecvMsg(&chStart[-(overlap+BL_VSIZE)],lowerRank);
				blWaitMsgDMA(lowerRank);
			}
			if(sendUp)
				blWaitMsgAck(upperRank);
			
			if(sendDown) // send edge downwards
				blSendMsg(baX.current, (overlap+BL_VSIZE)*sizeof(BL_DTYPE),lowerRank);
			if(recvFromUp)
			{
				blRecvMsg(lastBuf, upperRank);
				blWaitMsgDMA(upperRank);
			}
			if(sendDown)
				blWaitMsgAck(lowerRank);
		}
		else if(edge == BL_CYCLIC)
		{
			// One spu in group and cyclic mode, rare case
			// slow and ugly, copied from above code
			int alignOff;
			if(baX.part%BL_VSIZE)
				alignOff = BL_VSIZE-baX.part%BL_VSIZE;
			else
				alignOff = 0;
			
			get_BArr_dA(x, &chOverStart[BL_CHSIZE - (overlap+BL_VSIZE)], baX.start + baX.part - (overlap+BL_VSIZE) + alignOff, baX.start + baX.part -1);
			wait_BArr_complete(x, (Addr64)&chOverStart[BL_CHSIZE - (overlap+BL_VSIZE)]);
			copy(&chOverStart[BL_CHSIZE - (overlap+BL_VSIZE+alignOff)], &chStart[-(overlap+BL_VSIZE)], overlap+BL_VSIZE);
			copy(baX.current, lastBuf, overlap);
		}
		
		if(rank == 0 && edge == BL_ZERO)
		{
			zero(&chStart[-overlap-BL_VSIZE], overlap+BL_VSIZE);
		}
		stopTimer(&bl_time.comb);
		
#else
		// NestStep combine version
		startTimer(&bl_time.comb);
		// prime edges
		// lower edge
		if(rank != 0) // copy from left?
		{
			read_request_b(lowMirrorBuf, x, baX.start - overlap, baX.start-1);
		}
		else if(edge == BL_CYCLIC)
		{
			read_request_b(lowMirrorBuf, x, N-overlap, N-1);
		}
		// upper edge
		if(rank != groupSize-1)
		{
			read_request_b(highMirrorBuf, x, baX.start + baX.part, baX.start + baX.part + overlap -1);
		}
		else if(edge == BL_CYCLIC)
			read_request_b(highMirrorBuf, x, 0, overlap-1);
	
		// the combine will read request (slow!)
		NestStep_combine(NULL,NULL);
		stopTimer(&bl_time.comb);
	}
	NestStep_end_step();

	NestStep_step();
	{
		// read the mirrored data from memory
		startTimer(&bl_time.comb);
		if(rank != 0 || edge == BL_CYCLIC)
		{
			get_PArr(lowMirrorBuf,(Addr64)&(chStart[-overlap]),0,overlap-1);
			wait_PArr_complete(lowMirrorBuf, (Addr64)&(chStart[-overlap]));
		}
		else
			zero(&chStart[-overlap], overlap);
	
		if(rank != groupSize-1 || edge == BL_CYCLIC)
		{
			get_PArr(highMirrorBuf,(Addr64)lastBuf,0,overlap-1);
		}
		stopTimer(&bl_time.comb);
		get_sw_BArr(&baX);
#endif
		
		copy(baX.current, chStart, baX.currentSize);
		int lastCurrentSize = baX.currentSize;
		int lastStart = baX.currentStart;
		int diff = 0;
		// This skeleton does too much copying. Some of them could be avoided. 
		while(get_sw_BArr(&baX)) // get chunks
		{
			// copy new to overlap
			copy(baX.current, &chStart[lastCurrentSize], overlap);
			if(overlap > baX.currentSize) // if last chunk is smaller than overlap
			{
				diff = overlap - baX.currentSize;
				if(rank != groupSize-1 || edge == BL_CYCLIC)
				{
					// copy from left
#ifndef BL_MP_COMB
					wait_PArr_complete(highMirrorBuf, (Addr64)lastBuf);
#endif
					copy(lastBuf, &chStart[lastCurrentSize+baX.currentSize], overlap-baX.currentSize);
				}
				else
					zero(&chStart[lastCurrentSize+baX.currentSize], overlap-baX.currentSize); 
			}			
			
			// perform the actual work
			startTimer(&bl_time.calc);
			map_func(chStart, baRes.current, lastStart, lastCurrentSize);
			stopTimer(&bl_time.calc);
			put_sw_BArr(&baRes); // put res back
			
			// copy overlap to the left
			copy(&chStart[lastCurrentSize - overlap], &chStart[-overlap], overlap);
			copy(baX.current, chStart, baX.currentSize);
			lastCurrentSize = baX.currentSize;
			lastStart = baX.currentStart;
		}
		
		//fix last edge
		if(rank != groupSize-1 || edge == BL_CYCLIC)
		{
			#ifndef BL_MP_COMB	
			if(diff == 0)
				wait_PArr_complete(highMirrorBuf, (Addr64)lastBuf);
			#endif
			copy(lastBuf, &chStart[lastCurrentSize], overlap);
		}
		else if(edge == BL_ZERO)
			zero(&chStart[lastCurrentSize], overlap); 
		
		startTimer(&bl_time.calc);
		map_func(chStart, baRes.current, lastStart, lastCurrentSize);
		stopTimer(&bl_time.calc);
		put_sw_BArr(&baRes);
		wait_put_BArr_done(&baRes); // wait for last put
#ifdef BL_MP_COMB	
		blMPGroupSync();
#else
		startTimer(&bl_time.comb);
		NestStep_combine(NULL,NULL);
		stopTimer(&bl_time.comb);
#endif
	}
	NestStep_end_step();
	stopTimer(&bl_time.total);
}


static BL_DTYPE (*globalOverlappedMapFunc)(BL_DTYPE*, int) = NULL;

static void localOverlappedMapGlobalFunc(BL_DTYPE *x, BL_DTYPE *res, int startN, int n)
{
	int i;
	for(i=0;i<n;i++)
		res[i] = globalOverlappedMapFunc(&x[i], startN + i);
}

// wraper for function pointer
void BL_TN(DistOverlappedMap)(BL_DTYPE (*map_func)(BL_DTYPE*, int), BlockDistArray* x, BlockDistArray* res, int N, int overlap, enum bl_edge_value edge)
{
	globalOverlappedMapFunc = map_func;
	BL_TN(DistOverlappedMapLocalFunc)(&localOverlappedMapGlobalFunc, x, res, N, overlap, edge);
}
