#ifndef __SPU_BLOCK_MAP_H__
#define __SPU_BLOCK_MAP_H__

#include <NestStep_spulib/neststep_spu.h>
#include "vector_defs.h"
#include "bl_common.h"

#define MAX_SPUS 8
#define NAME_OFFSET 1024
#define NAME_OVERLAP 5 
void sDistOverlappedMapLocalFunc(void (*map_func)(float*,float*,int,int), BlockDistArray* x, BlockDistArray* res, int N, int overlap, enum bl_edge_value edge);
void sDistOverlappedMap(float (*map_func)(float*, int), BlockDistArray* x, BlockDistArray* res, int N, int overlap, enum bl_edge_value edge);

void dDistOverlappedMapLocalFunc(void (*map_func)(double*,double*,int,int), BlockDistArray* x, BlockDistArray* res, int N, int overlap, enum bl_edge_value edge);
void dDistOverlappedMap(double (*map_func)(double*, int), BlockDistArray* x, BlockDistArray* res, int N, int overlap, enum bl_edge_value edge);

#define BL_SINDEXED(name, offset) vector float name __attribute__((aligned(16))) = {x[offset],x[offset+1],x[offset+2],x[offset+3]};
#define BL_DINDEXED(name, offset) vector double name __attribute__((aligned(16))) = {x[offset],x[offset+1]};

#define DEF_OVERLAPPED_MAP_ONE_FUNC_S(name, result, overlap, consts, funcs)\
/* */\
float name##_2op(float *x, int bl_start)\
	{\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(float *x, float *res, int bl_start, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		{\
			consts \
			nVec = n/4;\
			for(i=0;i<nVec;i++)\
			{\
				funcs\
				((vector float*)res)[i] = result;\
				x+=4;\
			}\
			x-=nVec*4;\
		}\
		for(i=nVec*4;i<n;i++)\
		{\
			res[i] = name##_2op(&x[i], bl_start+i);\
		}\
	}\
	\
	void name(BlockDistArray *x, BlockDistArray *res,  int n, enum bl_edge_value edge)\
	{\
		sDistOverlappedMapLocalFunc(&name##_local,x, res,n, overlap, edge);\
	}

#define DEF_OVERLAPPED_MAP_ONE_FUNC_D(name, result, overlap, consts, funcs)\
/* */\
double name##_2op(double *x, int bl_start)\
	{\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(double *x, double *res, int bl_start, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		{\
			consts \
			nVec = n/2;\
			for(i=0;i<nVec;i++)\
			{\
				funcs\
				((vector double*)res)[i] = result;\
				x+=2;\
			}\
			x-=nVec*2;\
		}\
		for(i=nVec*2;i<n;i++)\
		{\
			res[i] = name##_2op(&x[i], bl_start+i);\
		}\
	}\
	\
	void name(BlockDistArray *x, BlockDistArray *res, int n, enum bl_edge_value edge)\
	{\
		dDistOverlappedMapLocalFunc(&name##_local,x, res, n, overlap, edge);\
	}

#endif
