#ifndef __SPU_GENERATE_H__
#define __SPU_GENERATE_H__
#include <NestStep_spulib/neststep_spu.h>
#include "vector_defs.h"
#include "bl_common.h"
#define NAME_OFFSET 1024
#define NAME_GENERATE 4
#define MAX_SPUS 8

void sDistGenerate(float (*gen_func)(int), BlockDistArray* x, int N, enum bl_do_step do_step);
void sDistGenerateLocalFunc(void (*gen_func)(float*, int,int), BlockDistArray* x, int N, enum bl_do_step do_step);

void dDistGenerate(double (*gen_func)(int), BlockDistArray* x, int N, enum bl_do_step do_step);
void dDistGenerateLocalFunc(void (*gen_func)(double*, int,int), BlockDistArray* x, int N, enum bl_do_step do_step);

void sPrivateGenerate(float (*gen_func)(int), PrivateArray* x, int N, enum bl_do_step do_step);
void sPrivateGenerateLocalFunc(void (*gen_func)(float*, int,int), PrivateArray* x, int N, enum bl_do_step do_step);

void dPrivateGenerate(double (*gen_func)(int), PrivateArray* x, int N, enum bl_do_step do_step);
void dPrivateGenerateLocalFunc(void (*gen_func)(double*, int,int), PrivateArray* x, int N, enum bl_do_step do_step);

#define DEF_GENERATE_FUNC_S(name, result, constants, funcs)\
	DEF_GENERATE_FUNC_TYPE_S(Dist, BlockDist, name, result, constants, funcs)

#define DEF_GENERATE_FUNC_PRIVATE_S(name, result, constants, funcs)\
	DEF_GENERATE_FUNC_TYPE_S(Private, Private, name, result, constants, funcs)

#define DEF_GENERATE_FUNC_DIST_S(name, result, constants, funcs)\
	DEF_GENERATE_FUNC_TYPE_S(Dist, BlockDist, name, result, constants, funcs)

#define DEF_GENERATE_FUNC_TYPE_S(type, ar_type, name, result, consts, funcs)\
/* */\
float name##_one(int iIs)\
	{\
		float iFs = (float) iIs;\
		vector float iF __attribute__((aligned(16))) = spu_splats(iFs);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(float *res, int start, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		{\
			consts \
			nVec = n/4;\
			vector float offset = {0,1,2,3};\
			vector float start_vec = spu_splats((float)start);\
			vector float four = {4,4,4,4};\
			for(i=0;i<nVec;i++)\
			{\
				vector float iF = spu_add(offset, start_vec);\
				offset = spu_add(offset, four);\
				consts \
				funcs \
				((vector float*)res)[i] = result;\
			}\
		}\
		for(i=nVec*4;i<n;i++)\
		{\
			res[i] = name##_one(i+start);\
		}\
	}\
	\
	void name(ar_type##Array *res, int n, enum bl_do_step do_step)\
	{\
		s##type##GenerateLocalFunc(&name##_local, res, n, do_step);\
	}

#define DEF_GENERATE_FUNC_D(name, result, constants, funcs)\
	DEF_GENERATE_FUNC_TYPE_D(Dist, BlockDist,  name, result, constants, funcs)

#define DEF_GENERATE_FUNC_PRIVATE_D(name, result, constants, funcs)\
	DEF_GENERATE_FUNC_TYPE_D(Private, Private, name, result, constants, funcs)

#define DEF_GENERATE_FUNC_DIST_D(name, result, constants, funcs)\
	DEF_GENERATE_FUNC_TYPE_D(Dist, BlockDist, name, result, constants, funcs)
	
#define DEF_GENERATE_FUNC_TYPE_D(type, ar_type, name, result, consts, funcs)\
/* */\
double name##_one(int iIs)\
	{\
		double iFs = (double) iIs;\
		vector double iF __attribute__((aligned(16))) = spu_splats(iFs);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(double *res, int start, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=2)\
		{\
			consts \
			nVec = n/2;\
			vector double offset = {0,1};\
			vector double start_vec = spu_splats((double)start);\
			vector double four = {2,2};\
			for(i=0;i<nVec;i++)\
			{\
				vector double iF = spu_add(offset, start_vec);\
				offset = spu_add(offset, four);\
				consts \
				funcs \
				((vector double*)res)[i] = result;\
			}\
		}\
		for(i=nVec*2;i<n;i++)\
		{\
			res[i] = name##_one(i+start);\
		}\
	}\
	\
	void name(BlockDistArray *res, int n, enum bl_do_step do_step)\
	{\
		dDistGenerateLocalFunc(&name##_local, res, n, do_step);\
	}


#endif
