#ifndef __SPU_MAP_H__
#define __SPU_MAP_H__

#include <NestStep_spulib/neststep_spu.h>
#include "vector_defs.h"
#include "bl_common.h"
#define NAME_OFFSET 1024
#define NAME_MAP 2
#define MAX_SPUS 8

// map functions with one argument array
void distMapLocalFunc(void (*map_func)(float**,int,int), BlockDistArray** x, int nOperators, BlockDistArray* res, int N);

void sDistMapOne(float (*map_func)(float), BlockDistArray* x1, BlockDistArray* res, int N, enum bl_do_step do_step);
void sDistMapOneLocalFunc(void (*map_func)(float*,float*,int), BlockDistArray* x, BlockDistArray* res, int N, enum bl_do_step do_step);

void dDistMapOne(double (*map_func)(double), BlockDistArray* x1, BlockDistArray* res, int N, enum bl_do_step do_step);
void dDistMapOneLocalFunc(void (*map_func)(double*,double*,int), BlockDistArray* x, BlockDistArray* res, int N, enum bl_do_step do_step);

// map functions with rwo argument arrays
void sDistMapTwo(float (*map_func)(float,float), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* res, int N, enum bl_do_step do_step);
void sDistMapTwoLocalFunc(void (*map_func)(float*,float*,float*,int), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* res, int N, enum bl_do_step do_step);

void dDistMapTwo(double (*map_func)(double,double), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* res, int N, enum bl_do_step do_step);
void dDistMapTwoLocalFunc(void (*map_func)(double*,double*,double*,int), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* res, int N, enum bl_do_step do_step);

// map functions with three argument arrays
void sDistMapThree(float (*map_func)(float,float,float), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* x3, BlockDistArray* res, int N, enum bl_do_step do_step);
void sDistMapThreeLocalFunc(void (*map_func)(float*,float*,float*,float*,int), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* x3, BlockDistArray* res, int N, enum bl_do_step do_step);

void dDistMapThree(double (*map_func)(double,double,double), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* x3, BlockDistArray* res, int N, enum bl_do_step do_step);
void dDistMapThreeLocalFunc(void (*map_func)(double*,double*,double*,double*,int), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* x3, BlockDistArray* res, int N, enum bl_do_step do_step);

static inline float mapOne(void(*map_func)(float*,float*,float*),float a,float b)
{
	float tmp;
	map_func(&a,&b,&tmp);
	return tmp;
};

#define DEF_MAP_ONE_FUNC_S(name, result, consts, funcs)\
/* */\
float name##_2op(float opS)\
	{\
		vector float op __attribute__((aligned(16))) = spu_splats(opS);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(float *x, float *res, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		{\
			consts \
			nVec = n/4;\
			for(i=0;i<nVec;i++)\
			{\
				vector float op __attribute__((aligned(16))) = ((vector float*)x)[i];\
				funcs\
				((vector float*)res)[i] = result;\
			}\
		}\
		for(i=nVec*4;i<n;i++)\
		{\
			res[i] = name##_2op(x[i]);\
		}\
	}\
	\
	void name(BlockDistArray *x, BlockDistArray *res,  int n, enum  bl_do_step do_step)\
	{\
		sDistMapOneLocalFunc(&name##_local,x, res,n, do_step);\
	}


#define DEF_MAP_TWO_FUNC_S(name, result, consts, funcs)\
/* */\
float name##_2op(float op1S, float op2S)\
	{\
		vector float op1 __attribute__((aligned(16))) = spu_splats(op1S);\
		vector float op2 __attribute__((aligned(16))) = spu_splats(op2S);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(float *x1, float *x2, float *res, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		{\
			consts \
			nVec = n/4;\
			for(i=0;i<nVec;i++)\
			{\
				vector float op1 __attribute__((aligned(16))) = ((vector float*)x1)[i];\
				vector float op2 __attribute__((aligned(16))) = ((vector float*)x2)[i];\
				funcs\
				((vector float*)res)[i] = result;\
			}\
		}\
		for(i=nVec*4;i<n;i++)\
		{\
			res[i] = name##_2op(x1[i], x2[i]);\
		}\
	}\
	\
	void name(BlockDistArray *x1, BlockDistArray *x2, BlockDistArray *res,  int n, enum  bl_do_step do_step)\
	{\
		sDistMapTwoLocalFunc(&name##_local, x1, x2, res, n, do_step);\
	}

#define DEF_MAP_THREE_FUNC_S(name, result, consts, funcs)\
/* */\
float name##_2op(float op1S, float op2S, float op3S)\
	{\
		vector float op1 __attribute__((aligned(16))) = spu_splats(op1S);\
		vector float op2 __attribute__((aligned(16))) = spu_splats(op2S);\
		vector float op3 __attribute__((aligned(16))) = spu_splats(op3S);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(float *x1, float *x2, float *x3, float *res, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		__align_hint(x1, 16,0);\
		__align_hint(x2, 16,0);\
		__align_hint(x3, 16,0);\
		__align_hint(res, 16,0);\
		consts \
		{\
			nVec = n/4;\
			for(i=0;i<nVec;i++)\
			{\
				vector float op1 __attribute__((aligned(16))) = ((vector float*)x1)[i];\
				vector float op2 __attribute__((aligned(16))) = ((vector float*)x2)[i];\
				vector float op3 __attribute__((aligned(16))) = ((vector float*)x3)[i];\
				funcs\
				((vector float*)res)[i] = result;\
			}\
		}\
		for(i=nVec*4;i<n;i++)\
		{\
			res[i] = name##_2op(x1[i], x2[i], x3[i]);\
		}\
	}\
	\
	void name(BlockDistArray *x1, BlockDistArray *x2, BlockDistArray *x3, BlockDistArray *res,  int n, enum  bl_do_step do_step)\
	{\
		sDistMapThreeLocalFunc(&name##_local, x1, x2, x3, res, n, do_step);\
	}
	

#define DEF_MAP_ONE_FUNC_D(name, result, consts, funcs)\
/* */\
double name##_2op(double opS)\
	{\
		vector double op __attribute__((aligned(16))) = spu_splats(opS);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(double *x, double *res, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=2)\
		{\
			consts \
			nVec = n/2;\
			for(i=0;i<nVec;i++)\
			{\
				vector double op __attribute__((aligned(16))) = ((vector double*)x)[i];\
				funcs\
				((vector double*)res)[i] = result;\
			}\
		}\
		for(i=nVec*2;i<n;i++)\
		{\
			res[i] = name##_2op(x[i]);\
		}\
	}\
	\
	void name(BlockDistArray *x, BlockDistArray *res,  int n, enum  bl_do_step do_step)\
	{\
		dDistMapOneLocalFunc(&name##_local,x, res,n, do_step);\
	}


#define DEF_MAP_TWO_FUNC_D(name, result, consts, funcs)\
/* */\
double name##_2op(double op1S, double op2S)\
	{\
		vector double op1 __attribute__((aligned(16))) = spu_splats(op1S);\
		vector double op2 __attribute__((aligned(16))) = spu_splats(op2S);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(double *x1, double *x2, double *res, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=2)\
		{\
			consts \
			nVec = n/2;\
			for(i=0;i<nVec;i++)\
			{\
				vector double op1 __attribute__((aligned(16))) = ((vector double*)x1)[i];\
				vector double op2 __attribute__((aligned(16))) = ((vector double*)x2)[i];\
				funcs\
				((vector double*)res)[i] = result;\
			}\
		}\
		for(i=nVec*2;i<n;i++)\
		{\
			res[i] = name##_2op(x1[i], x2[i]);\
		}\
	}\
	\
	void name(BlockDistArray *x1, BlockDistArray *x2, BlockDistArray *res,  int n, enum  bl_do_step do_step)\
	{\
		dDistMapTwoLocalFunc(&name##_local, x1, x2, res, n, do_step);\
	}

#define DEF_MAP_THREE_FUNC_D(name, result, consts, funcs)\
/* */\
double name##_2op(double op1S, double op2S, double op3S)\
	{\
		vector double op1 __attribute__((aligned(16))) = spu_splats(op1S);\
		vector double op2 __attribute__((aligned(16))) = spu_splats(op2S);\
		vector double op3 __attribute__((aligned(16))) = spu_splats(op3S);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	void name##_local(double *x1, double *x2, double *x3, double *res, int n)\
	{\
		int i;\
		int nVec = 0;\
		if(n>=2)\
		__align_hint(x1, 16,0);\
		__align_hint(x2, 16,0);\
		__align_hint(x3, 16,0);\
		__align_hint(res, 16,0);\
		consts \
		{\
			nVec = n/2;\
			for(i=0;i<nVec;i++)\
			{\
				vector double op1 __attribute__((aligned(16))) = ((vector double*)x1)[i];\
				vector double op2 __attribute__((aligned(16))) = ((vector double*)x2)[i];\
				vector double op3 __attribute__((aligned(16))) = ((vector double*)x3)[i];\
				funcs\
				((vector double*)res)[i] = result;\
			}\
		}\
		for(i=nVec*2;i<n;i++)\
		{\
			res[i] = name##_2op(x1[i], x2[i], x3[i]);\
		}\
	}\
	\
	void name(BlockDistArray *x1, BlockDistArray *x2, BlockDistArray *x3, BlockDistArray *res,  int n, enum  bl_do_step do_step)\
	{\
		dDistMapThreeLocalFunc(&name##_local, x1, x2, x3, res, n, do_step);\
	}
	
	
#endif	
