#ifndef __SPU_MAP_REDUCE_H__
#define __SPU_MAP_REDUCE_H__

#include <NestStep_spulib/neststep_spu.h>
#include "vector_defs.h"
#define NAME_OFFSET 1024
#define NAME_MAP_REDUCE 3
#define MAX_SPUS 8


// map-reduce functions with one argument array
float sDistMapOneAndReduceLocalFunc(float (*map_red_func)(float*,int), float (*red_func)(float*,int),BlockDistArray* x, int N);
double dDistMapOneAndReduceLocalFunc(double (*map_red_func)(double*,int), double (*red_func)(double*,int),BlockDistArray* x, int N);
// map-reduce functions with two argument arrays
float sDistMapTwoAndReduceLocalFunc(float (*map_red_func)(float*,float*,int), float (*red_func)(float*,int),BlockDistArray* x1,BlockDistArray* x2, int N);
double dDistMapTwoAndReduceLocalFunc(double (*map_red_func)(double*,double*,int), double (*red_func)(double*,int),BlockDistArray* x1,BlockDistArray* x2, int N);
// map-reduce functions with three argument arrays
float sDistMapThreeAndReduceLocalFunc(float (*map_red_func)(float*,float*,float*,int), float (*red_func)(float*,int),BlockDistArray* x1,BlockDistArray* x2,BlockDistArray* x3, int N);
double dDistMapThreeAndReduceLocalFunc(double (*map_red_func)(double*,double*,double*,int), double (*red_func)(double*,int),BlockDistArray* x1,BlockDistArray* x2,BlockDistArray* x3, int N);


#define DEF_MAP_ONE_AND_REDUCE_FUNC_S(name, map_result, red_result, bl_consts, map_funcs, red_funcs)\
/* */\
float name##_one(float m_op1S, float r_op2S)\
	{\
		vector float m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector float r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		map_funcs \
		vector float r_op1 __attribute__((aligned(16))) = map_result;\
		red_funcs \
		return spu_extract(red_result,0);\
	}\
	\
float name##_one_ro(float r_op1S, float r_op2S)\
	{\
		vector float r_op1 __attribute__((aligned(16))) = spu_splats(r_op1S);\
		vector float r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		red_funcs\
		return spu_extract(red_result,0);\
	}\
float name##_one_mo(float m_op1S)\
	{\
		vector float m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		bl_consts \
		map_funcs\
		return spu_extract(map_result,0);\
	}\
	\
	float name##_local(float *x1, int n)\
	{\
		float res;\
		int i;\
		int nVec = 0;\
		if(n>=8)\
		{\
			bl_consts \
			nVec = n/4;\
			vector float vec_res __attribute__((aligned(16)));\
			{\
				vector float m_op1 __attribute__((aligned(16))) = ((vector float*)x1)[0];\
				map_funcs \
				vec_res = map_result;\
			}\
			for(i=1;i<nVec;i++)\
			{\
				vector float m_op1 __attribute__((aligned(16))) = ((vector float*)x1)[i];\
				map_funcs\
				vector float r_op2 __attribute__((aligned(16))) = vec_res;\
				vector float r_op1 __attribute__((aligned(16))) = map_result;\
				red_funcs \
				vec_res = red_result;\
			}\
			res = name##_one_ro(name##_one_ro(spu_extract(vec_res, 0), spu_extract(vec_res, 1)),\
						name##_one_ro(spu_extract(vec_res, 2), spu_extract(vec_res, 3)));\
		}\
		if(nVec*4 == n)\
		{\
			return res;\
		}\
		if(nVec)\
			res = name##_one(x1[nVec*4], res);\
		else \
			res = name##_one_mo(x1[0]);\
		for(i=nVec*4+1;i<n;i++)\
		{\
			res = name##_one(x1[i],res);\
		}\
		return res;\
	}\
	float name##_local_ro(float *x1, int n)\
	{\
		int i;\
		float res = x1[0];\
		for(i=1;i<n;i++)\
		{\
			res = name##_one_ro(x1[i],res);\
		}\
		return res;\
	}\
	\
	float name(BlockDistArray *x1,int n)\
	{\
		return sDistMapOneAndReduceLocalFunc(&name##_local,&name##_local_ro,x1,n);\
	}


#define DEF_MAP_TWO_AND_REDUCE_FUNC_S(name, map_result, red_result, bl_consts, map_funcs, red_funcs)\
/* */\
float name##_one(float m_op1S, float m_op2S, float r_op2S)\
	{\
		vector float m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector float m_op2 __attribute__((aligned(16))) = spu_splats(m_op2S);\
		vector float r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		map_funcs \
		vector float r_op1 __attribute__((aligned(16))) = map_result;\
		red_funcs \
		return spu_extract(red_result,0);\
	}\
	\
float name##_one_ro(float r_op1S, float r_op2S)\
	{\
		vector float r_op1 __attribute__((aligned(16))) = spu_splats(r_op1S);\
		vector float r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		red_funcs\
		return spu_extract(red_result,0);\
	}\
float name##_one_mo(float m_op1S, float m_op2S)\
	{\
		vector float m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector float m_op2 __attribute__((aligned(16))) = spu_splats(m_op2S);\
		bl_consts \
		map_funcs\
		return spu_extract(map_result,0);\
	}\
	\
	float name##_local(float *x1, float *x2, int n)\
	{\
		float res;\
		int i;\
		int nVec = 0;\
		if(n>=8)\
		{\
			bl_consts \
			nVec = n/4;\
			vector float vec_res __attribute__((aligned(16)));\
			{\
				vector float m_op1 __attribute__((aligned(16))) = ((vector float*)x1)[0];\
				vector float m_op2 __attribute__((aligned(16))) = ((vector float*)x2)[0];\
				map_funcs \
				vec_res = map_result;\
			}\
			for(i=1;i<nVec;i++)\
			{\
				vector float m_op1 __attribute__((aligned(16))) = ((vector float*)x1)[i];\
				vector float m_op2 __attribute__((aligned(16))) = ((vector float*)x2)[i];\
				map_funcs\
				vector float r_op2 __attribute__((aligned(16))) = vec_res;\
				vector float r_op1 __attribute__((aligned(16))) = map_result;\
				red_funcs \
				vec_res = red_result;\
			}\
			res = name##_one_ro(name##_one_ro(spu_extract(vec_res, 0), spu_extract(vec_res, 1)),\
						name##_one_ro(spu_extract(vec_res, 2), spu_extract(vec_res, 3)));\
		}\
		if(nVec*4 == n)\
		{\
			return res;\
		}\
		if(nVec)\
			res = name##_one(x1[nVec*4], x2[nVec*4],res);\
		else \
			res = name##_one_mo(x1[0], x2[0]);\
		for(i=nVec*4+1;i<n;i++)\
		{\
			res = name##_one(x1[i],x2[i],res);\
		}\
		return res;\
	}\
	float name##_local_ro(float *x1, int n)\
	{\
		int i;\
		float res = x1[0];\
		for(i=1;i<n;i++)\
		{\
			res = name##_one_ro(x1[i],res);\
		}\
		return res;\
	}\
	\
	float name(BlockDistArray *x1,BlockDistArray *x2, int n)\
	{\
		return sDistMapTwoAndReduceLocalFunc(&name##_local,&name##_local_ro,x1,x2,n);\
	}
	

#define DEF_MAP_THREE_AND_REDUCE_FUNC_S(name, map_result, red_result, bl_consts, map_funcs, red_funcs)\
/* */\
float name##_one(float m_op1S, float m_op2S, float m_op3S, float r_op2S)\
	{\
		vector float m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector float m_op2 __attribute__((aligned(16))) = spu_splats(m_op2S);\
		vector float m_op3 __attribute__((aligned(16))) = spu_splats(m_op3S);\
		vector float r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		map_funcs \
		vector float r_op1 __attribute__((aligned(16))) = map_result;\
		red_funcs \
		return spu_extract(red_result,0);\
	}\
	\
float name##_one_ro(float r_op1S, float r_op2S)\
	{\
		vector float r_op1 __attribute__((aligned(16))) = spu_splats(r_op1S);\
		vector float r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		red_funcs\
		return spu_extract(red_result,0);\
	}\
float name##_one_mo(float m_op1S, float m_op2S, float m_op3S)\
	{\
		vector float m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector float m_op2 __attribute__((aligned(16))) = spu_splats(m_op2S);\
		vector float m_op3 __attribute__((aligned(16))) = spu_splats(m_op3S);\
		bl_consts \
		map_funcs\
		return spu_extract(map_result,0);\
	}\
	\
	float name##_local(float *x1, float *x2, float *x3, int n)\
	{\
		float res;\
		int i;\
		int nVec = 0;\
		if(n>=8)\
		{\
			bl_consts \
			nVec = n/4;\
			vector float vec_res __attribute__((aligned(16)));\
			{\
				vector float m_op1 __attribute__((aligned(16))) = ((vector float*)x1)[0];\
				vector float m_op2 __attribute__((aligned(16))) = ((vector float*)x2)[0];\
				vector float m_op3 __attribute__((aligned(16))) = ((vector float*)x3)[0];\
				map_funcs \
				vec_res = map_result;\
			}\
			for(i=1;i<nVec;i++)\
			{\
				vector float m_op1 __attribute__((aligned(16))) = ((vector float*)x1)[i];\
				vector float m_op2 __attribute__((aligned(16))) = ((vector float*)x2)[i];\
				vector float m_op3 __attribute__((aligned(16))) = ((vector float*)x3)[i];\
				map_funcs\
				vector float r_op2 __attribute__((aligned(16))) = vec_res;\
				vector float r_op1 __attribute__((aligned(16))) = map_result;\
				red_funcs \
				vec_res = red_result;\
			}\
			res = name##_one_ro(name##_one_ro(spu_extract(vec_res, 0), spu_extract(vec_res, 1)),\
						name##_one_ro(spu_extract(vec_res, 2), spu_extract(vec_res, 3)));\
		}\
		if(nVec*4 == n)\
		{\
			return res;\
		}\
		if(nVec)\
			res = name##_one(x1[nVec*4], x2[nVec*4], x3[nVec*4],res);\
		else \
			res = name##_one_mo(x1[0], x2[0], x3[0]);\
		for(i=nVec*4+1;i<n;i++)\
		{\
			res = name##_one(x1[i],x2[i],x3[i],res);\
		}\
		return res;\
	}\
	float name##_local_ro(float *x1, int n)\
	{\
		int i;\
		float res = x1[0];\
		for(i=1;i<n;i++)\
		{\
			res = name##_one_ro(x1[i],res);\
		}\
		return res;\
	}\
	\
	float name(BlockDistArray *x1, BlockDistArray *x2, BlockDistArray *x3, int n)\
	{\
		return sDistMapThreeAndReduceLocalFunc(&name##_local, &name##_local_ro, x1, x2, x3, n);\
	}
	

#define DEF_MAP_ONE_AND_REDUCE_FUNC_D(name, map_result, red_result, bl_consts, map_funcs, red_funcs)\
/* */\
double name##_one(double m_op1S, double r_op2S)\
	{\
		vector double m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector double r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		map_funcs \
		vector double r_op1 __attribute__((aligned(16))) = map_result;\
		red_funcs \
		return spu_extract(red_result,0);\
	}\
	\
double name##_one_ro(double r_op1S, double r_op2S)\
	{\
		vector double r_op1 __attribute__((aligned(16))) = spu_splats(r_op1S);\
		vector double r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		red_funcs\
		return spu_extract(red_result,0);\
	}\
double name##_one_mo(double m_op1S)\
	{\
		vector double m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		bl_consts \
		map_funcs\
		return spu_extract(map_result,0);\
	}\
	\
	double name##_local(double *x1, int n)\
	{\
		double res;\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		{\
			bl_consts \
			nVec = n/2;\
			vector double vec_res __attribute__((aligned(16)));\
			{\
				vector double m_op1 __attribute__((aligned(16))) = ((vector double*)x1)[0];\
				map_funcs \
				vec_res = map_result;\
			}\
			for(i=1;i<nVec;i++)\
			{\
				vector double m_op1 __attribute__((aligned(16))) = ((vector double*)x1)[i];\
				map_funcs\
				vector double r_op2 __attribute__((aligned(16))) = vec_res;\
				vector double r_op1 __attribute__((aligned(16))) = map_result;\
				red_funcs \
				vec_res = red_result;\
			}\
			res = name##_one_ro(spu_extract(vec_res, 0), spu_extract(vec_res, 1));\
		}\
		if(nVec*2 == n)\
		{\
			return res;\
		}\
		if(nVec)\
			res = name##_one(x1[nVec*2], res);\
		else \
			res = name##_one_mo(x1[0]);\
		for(i=nVec*2+1;i<n;i++)\
		{\
			res = name##_one(x1[i],res);\
		}\
		return res;\
	}\
	double name##_local_ro(double *x1, int n)\
	{\
		int i;\
		double res = x1[0];\
		for(i=1;i<n;i++)\
		{\
			res = name##_one_ro(x1[i],res);\
		}\
		return res;\
	}\
	\
	double name(BlockDistArray *x1,int n)\
	{\
		return dDistMapOneAndReduceLocalFunc(&name##_local,&name##_local_ro,x1,n);\
	}


#define DEF_MAP_TWO_AND_REDUCE_FUNC_D(name, map_result, red_result, bl_consts, map_funcs, red_funcs)\
/* */\
double name##_one(double m_op1S, double m_op2S, double r_op2S)\
	{\
		vector double m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector double m_op2 __attribute__((aligned(16))) = spu_splats(m_op2S);\
		vector double r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		map_funcs \
		vector double r_op1 __attribute__((aligned(16))) = map_result;\
		red_funcs \
		return spu_extract(red_result,0);\
	}\
	\
double name##_one_ro(double r_op1S, double r_op2S)\
	{\
		vector double r_op1 __attribute__((aligned(16))) = spu_splats(r_op1S);\
		vector double r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		red_funcs\
		return spu_extract(red_result,0);\
	}\
double name##_one_mo(double m_op1S, double m_op2S)\
	{\
		vector double m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector double m_op2 __attribute__((aligned(16))) = spu_splats(m_op2S);\
		bl_consts \
		map_funcs\
		return spu_extract(map_result,0);\
	}\
	\
	double name##_local(double *x1, double *x2, int n)\
	{\
		double res;\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		{\
			bl_consts \
			nVec = n/2;\
			vector double vec_res __attribute__((aligned(16)));\
			{\
				vector double m_op1 __attribute__((aligned(16))) = ((vector double*)x1)[0];\
				vector double m_op2 __attribute__((aligned(16))) = ((vector double*)x2)[0];\
				map_funcs \
				vec_res = map_result;\
			}\
			for(i=1;i<nVec;i++)\
			{\
				vector double m_op1 __attribute__((aligned(16))) = ((vector double*)x1)[i];\
				vector double m_op2 __attribute__((aligned(16))) = ((vector double*)x2)[i];\
				map_funcs\
				vector double r_op2 __attribute__((aligned(16))) = vec_res;\
				vector double r_op1 __attribute__((aligned(16))) = map_result;\
				red_funcs \
				vec_res = red_result;\
			}\
			res = name##_one_ro(spu_extract(vec_res, 0), spu_extract(vec_res, 1));\
		}\
		if(nVec*2 == n)\
		{\
			return res;\
		}\
		if(nVec)\
			res = name##_one(x1[nVec*2], x2[nVec*2],res);\
		else \
			res = name##_one_mo(x1[0], x2[0]);\
		for(i=nVec*2+1;i<n;i++)\
		{\
			res = name##_one(x1[i],x2[i],res);\
		}\
		return res;\
	}\
	double name##_local_ro(double *x1, int n)\
	{\
		int i;\
		double res = x1[0];\
		for(i=1;i<n;i++)\
		{\
			res = name##_one_ro(x1[i],res);\
		}\
		return res;\
	}\
	\
	double name(BlockDistArray *x1,BlockDistArray *x2, int n)\
	{\
		return dDistMapTwoAndReduceLocalFunc(&name##_local,&name##_local_ro,x1,x2,n);\
	}
	

#define DEF_MAP_THREE_AND_REDUCE_FUNC_D(name, map_result, red_result, bl_consts, map_funcs, red_funcs)\
/* */\
double name##_one(double m_op1S, double m_op2S, double m_op3S, double r_op2S)\
	{\
		vector double m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector double m_op2 __attribute__((aligned(16))) = spu_splats(m_op2S);\
		vector double m_op3 __attribute__((aligned(16))) = spu_splats(m_op3S);\
		vector double r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		map_funcs \
		vector double r_op1 __attribute__((aligned(16))) = map_result;\
		red_funcs \
		return spu_extract(red_result,0);\
	}\
	\
double name##_one_ro(double r_op1S, double r_op2S)\
	{\
		vector double r_op1 __attribute__((aligned(16))) = spu_splats(r_op1S);\
		vector double r_op2 __attribute__((aligned(16))) = spu_splats(r_op2S);\
		bl_consts \
		red_funcs\
		return spu_extract(red_result,0);\
	}\
double name##_one_mo(double m_op1S, double m_op2S, double m_op3S)\
	{\
		vector double m_op1 __attribute__((aligned(16))) = spu_splats(m_op1S);\
		vector double m_op2 __attribute__((aligned(16))) = spu_splats(m_op2S);\
		vector double m_op3 __attribute__((aligned(16))) = spu_splats(m_op3S);\
		bl_consts \
		map_funcs\
		return spu_extract(map_result,0);\
	}\
	\
	double name##_local(double *x1, double *x2, double *x3, int n)\
	{\
		double res;\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		{\
			bl_consts \
			nVec = n/2;\
			vector double vec_res __attribute__((aligned(16)));\
			{\
				vector double m_op1 __attribute__((aligned(16))) = ((vector double*)x1)[0];\
				vector double m_op2 __attribute__((aligned(16))) = ((vector double*)x2)[0];\
				vector double m_op3 __attribute__((aligned(16))) = ((vector double*)x3)[0];\
				map_funcs \
				vec_res = map_result;\
			}\
			for(i=1;i<nVec;i++)\
			{\
				vector double m_op1 __attribute__((aligned(16))) = ((vector double*)x1)[i];\
				vector double m_op2 __attribute__((aligned(16))) = ((vector double*)x2)[i];\
				vector double m_op3 __attribute__((aligned(16))) = ((vector double*)x3)[i];\
				map_funcs\
				vector double r_op2 __attribute__((aligned(16))) = vec_res;\
				vector double r_op1 __attribute__((aligned(16))) = map_result;\
				red_funcs \
				vec_res = red_result;\
			}\
			res = name##_one_ro(spu_extract(vec_res, 0), spu_extract(vec_res, 1));\
		}\
		if(nVec*2 == n)\
		{\
			return res;\
		}\
		if(nVec)\
			res = name##_one(x1[nVec*2], x2[nVec*2], x3[nVec*2],res);\
		else \
			res = name##_one_mo(x1[0], x2[0], x3[0]);\
		for(i=nVec*2+1;i<n;i++)\
		{\
			res = name##_one(x1[i],x2[i],x3[i],res);\
		}\
		return res;\
	}\
	double name##_local_ro(double *x1, int n)\
	{\
		int i;\
		double res = x1[0];\
		for(i=1;i<n;i++)\
		{\
			res = name##_one_ro(x1[i],res);\
		}\
		return res;\
	}\
	\
	double name(BlockDistArray *x1, BlockDistArray *x2, BlockDistArray *x3, int n)\
	{\
		return dDistMapThreeAndReduceLocalFunc(&name##_local, &name##_local_ro, x1, x2, x3, n);\
	}
	
#endif
