#ifndef __SPU_REDUCE_H__
#define __SPU_REDUCE_H__
#include <NestStep_spulib/neststep_spu.h>
#include "vector_defs.h"
#include "bl_common.h"
#define NAME_OFFSET 1024
#define NAME_REDUCE 1
#define MAX_SPUS 8

float sDistReduce(float (*red_func)(float, float), BlockDistArray* x, int N);
double dDistReduce(double (*red_func)(double, double), BlockDistArray* x, int N);
float sDistReduceLocalFunc(float (*red_func)(float*,int), BlockDistArray* x, int N);
double dDistReduceLocalFunc(double (*red_func)(double*,int), BlockDistArray* x, int N);

static inline float sReduceTwo(float(*red_func)(float*,int),float a,float b)
{
	float args[2]={a,b};
	return red_func(args,2);
};

static inline double dReduceTwo(double(*red_func)(double*,int),double a,double b)
{
	double args[2]={a,b};
	return red_func(args,2);
};

// visible externally for usage in map-reduce module
extern RepArray* b_bl_reduce_rep_arr;
extern RepArray* s_bl_reduce_rep_arr;
void dAllocReduceRepArray();
void sAllocReduceRepArray();



#define DEF_REDUCE_FUNC_S(name, result, consts, funcs)\
/* */\
float name##_2op(float op2S, float op1S)\
	{\
		vector float op1 __attribute__((aligned(16))) = spu_splats(op1S);\
		vector float op2 __attribute__((aligned(16)))= spu_splats(op2S);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	float name##_local(float *x, int n)\
	{\
		float res;\
		int i;\
		int nVec = 0;\
		if(n>=8)\
		{\
			consts \
			nVec = n/4;\
			vector float vec_res __attribute__((aligned(16)));\
			vec_res = ((vector float*)x)[0];\
			for(i=1;i<nVec;i++)\
			{\
				vector float op1 __attribute__((aligned(16))) = ((vector float*)x)[i];\
				vector float op2 __attribute__((aligned(16))) = vec_res;\
				funcs\
				vec_res = result;\
			}\
			res = name##_2op(name##_2op(spu_extract(vec_res, 0), spu_extract(vec_res, 1)),\
						name##_2op(spu_extract(vec_res, 2), spu_extract(vec_res, 3)));\
		}\
		if(nVec*4 == n)\
		{\
			return res;\
		}\
		if(nVec)\
			res = name##_2op(res,x[nVec*4]);\
		else \
			res = x[0];\
		for(i=nVec*4+1;i<n;i++)\
		{\
			res = name##_2op(res, x[i]);\
		}\
		return res;\
	}\
	\
	float name(BlockDistArray *x, int n)\
	{\
		return sDistReduceLocalFunc(&name##_local,x,n);\
	}

#define DEF_REDUCE_FUNC_D(name, result, consts, funcs)\
/* */\
double name##_2op(double op2S, double op1S)\
	{\
		vector double op1 __attribute__((aligned(16))) = spu_splats(op1S);\
		vector double op2 __attribute__((aligned(16)))= spu_splats(op2S);\
		consts \
		funcs \
		return spu_extract(result,0);\
	}\
	\
	double name##_local(double *x, int n)\
	{\
		double res;\
		int i;\
		int nVec = 0;\
		if(n>=4)\
		{\
			consts \
			nVec = n/2;\
			vector double vec_res __attribute__((aligned(16)));\
			vec_res = ((vector double*)x)[0];\
			for(i=1;i<nVec;i++)\
			{\
				vector double op1 __attribute__((aligned(16))) = ((vector double*)x)[i];\
				vector double op2 __attribute__((aligned(16))) = vec_res;\
				funcs\
				vec_res = result;\
			}\
			res = name##_2op(spu_extract(vec_res, 0), spu_extract(vec_res, 1));\
		}\
		if(nVec*2 == n)\
		{\
			return res;\
		}\
		if(nVec)\
			res = name##_2op(res,x[nVec*2]);\
		else \
			res = x[0];\
		for(i=nVec*2+1;i<n;i++)\
		{\
			res = name##_2op(res, x[i]);\
		}\
		return res;\
	}\
	\
	double name(BlockDistArray *x, int n)\
	{\
		return dDistReduceLocalFunc(&name##_local,x,n);\
	}
#endif
