#ifndef __SPU_VEC_FUNC_DEF_H__
#define __SPU_VEC_FUNC_DEF_H__

#include <simdmath.h>

// Macros for the function definition language. Expand as needed.

#define BL_SCONST(name, val) vector float name __attribute__((aligned(16))) = spu_splats(val);
#define BL_DCONST(name, val) vector double name __attribute__((aligned(16))) = spu_splats(val);
#define BL_NONE 

#define BL_SADD(name, o1,o2) vector float name = spu_add(o1, o2);
#define BL_DADD(name, o1,o2) vector double name = spu_add(o1, o2);
#define BL_SSUB(name, o1,o2) vector float name = spu_sub(o1, o2);
#define BL_DSUB(name, o1,o2) vector double name = spu_sub(o1, o2);
#define BL_SMUL(name, o1,o2) vector float name = spu_mul(o1, o2);
#define BL_DMUL(name, o1,o2) vector double name = spu_mul(o1, o2);
#define BL_SMIN(name, o1,o2) vector float name = (vector float)spu_sel(o1,o2, spu_cmpgt(o1, o2));
#define BL_DMIN(name, o1,o2) vector double name = (vector double)spu_sel(o1,o2, spu_cmpgt(o1, o2));
#define BL_SMAX(name, o1,o2) vector float name = (vector float)spu_sel(o1,o2, spu_cmpgt(o2, o1));
#define BL_DMAX(name, o1,o2) vector double name = (vector double)spu_sel(o1,o2, spu_cmpgt(o2, o1));

// common dual operations:
#define BL_SMUL_ADD(name, o1,o2,o3) vector float name = spu_madd(o1, o2, o3);
#define BL_DMUL_ADD(name, o1,o2,o3) vector double name = spu_madd(o1, o2, o3);
#define BL_SMUL_SUB(name, o1,o2,o3) vector float name = spu_msub(o1, o2, o3);
#define BL_DMUL_SUB(name, o1,o2,o3) vector double name = spu_msub(o1, o2, o3);

// ibm simdmath-lib
#define BL_SSQRT(name, o1) vector float name = sqrtf4(o1);
#define BL_DSQRT(name, o1) vector double name = sqrtd2(o1);
#define BL_SEXP(name, o1) vector float name =  expf4(o1);
#define BL_DEXP(name, o1) vector double name =  expd2(o1);
#define BL_SLOG(name, o1) vector float name =  logf4(o1);
#define BL_DLOG(name, o1) vector double name =  logd2(o1);
#define BL_SRECIP(name, o1) vector float name = recipf4(o1);
#define BL_DRECIP(name, o1) vector double name = recipd2(o1);
#define BL_SMOD(name, o1, o2) vector float name = modf4(o1,o2);
#define BL_DMOD(name, o1, o2) vector double name = modd2(o1,o2);
#define BL_SABS(name, o1) vector float name = fabsf4(o1);
#define BL_DABS(name, o1) vector double name = fabsd2(o1);


#define BL_SDIV(name, o1,o2) vector float name =  divf4(o1,o2);
#define BL_DDIV(name, o1,o2) vector double name =  divd2(o1,o2);


//Trigonometric
#define BL_SSIN(name, o1) vector float name = sinf4(o1);
#define BL_DSIN(name, o1) vector double name = sind2(o1);
#define BL_SCOS(name, o1) vector float name = cosf4(o1);
#define BL_DCOS(name, o1) vector double name = cosd2(o1);
#define BL_SASIN(name, o1) vector float name = asinf4(o1);
#define BL_DASIN(name, o1) vector double name = asind2(o1);
#define BL_SACOS(name, o1) vector float name = acosf4(o1);
#define BL_DACOS(name, o1) vector double name = acosd2(o1);
#define BL_STAN(name, o1) vector float name = tanf4(o1);
#define BL_DTAN(name, o1) vector double name = tand2(o1);
#define BL_SATAN(name, o1) vector float name = atanf4(o1);
#define BL_DATAN(name, o1) vector double name = atand2(o1);


#define DEF_TREE(name, func, o1, o2) \
	vector float name; \
	func(name,o1,o2)


#endif

