#include "include/map.h"
#include "include/spu_timer.h"
#include "include/spu_utils.h"
#include "include/bl_f_d.h"

// this file is compiled both for floats and doubles

void BL_TN(DistMapOneLocalFunc)(void (*map_func)(BL_DTYPE*,BL_DTYPE*,int), BlockDistArray* x1, BlockDistArray* res, int N, enum  bl_do_step do_step)
{
	startTimer(&bl_time.total);
	
	// init Barr handlers
	BArr_Handler baX;
	init_BArr_Get(&baX, x1, BL_CHSIZE);

	BArr_Handler baRes;
	init_BArr_Put(&baRes, res, BL_CHSIZE);
	
	if(do_step)
		NestStep_step();
    {	
		while(get_sw_BArr(&baX)) // get chunks
		{
			
			startTimer(&bl_time.calc);
			map_func(baX.current, baRes.current, baX.currentSize);
			stopTimer(&bl_time.calc);
			
			put_sw_BArr(&baRes); // put res back
		}

		wait_put_BArr_done(&baRes); // wait for last put
		
		if(do_step)
		{
#ifdef BL_MP_COMB
			blMPGroupSync();
#else
			startTimer(&bl_time.comb);
			NestStep_combine(NULL,NULL);
			stopTimer(&bl_time.comb);
#endif
		}
	}
	if(do_step)
		NestStep_end_step();
	stopTimer(&bl_time.total);
}


static BL_DTYPE (*globalMapOneFunc)(BL_DTYPE) = NULL;

static void localMapOneGlobalFunc(BL_DTYPE *x, BL_DTYPE *res, int n)
{
	int i;
	for(i=0;i<n;i++)
		res[i] = globalMapOneFunc(x[i]);
}

void BL_TN(DistMapOne)(BL_DTYPE (*map_func)(BL_DTYPE), BlockDistArray* x1, BlockDistArray* res, int N, enum  bl_do_step do_step)
{
	globalMapOneFunc = map_func;
	BL_TN(DistMapOneLocalFunc)(&localMapOneGlobalFunc, x1, res, N, do_step);
}

// The code below is just a copy of the one above but with more argument vectors. Could be rewritten as 
// one set of functions with some more macro hacks or with pure c-kod and some wrapers for 
// stack-allocation of buffers.
void BL_TN(DistMapTwoLocalFunc)(void (*map_func)(BL_DTYPE*,BL_DTYPE*,BL_DTYPE*,int), BlockDistArray* x1, BlockDistArray* x2, BlockDistArray* res, int N, enum bl_do_step do_step)
{
	startTimer(&bl_time.total);
	
	// init Barr handlers
	BArr_Handler baX1;
	init_BArr_Get(&baX1, x1, BL_CHSIZE);
	BArr_Handler baX2;
	init_BArr_Get(&baX2, x2, BL_CHSIZE);

	BArr_Handler baRes;
	init_BArr_Put(&baRes, res, BL_CHSIZE);
	
	if(do_step)
		NestStep_step();
    {	
		while(get_sw_BArr(&baX1)) // get chunks
		{
			get_sw_BArr(&baX2);
			
			startTimer(&bl_time.calc);
			map_func(baX1.current, baX2.current, baRes.current, baX1.currentSize);
			stopTimer(&bl_time.calc);
			
			put_sw_BArr(&baRes); // put res back
		}

		wait_put_BArr_done(&baRes); // wait for last put
		
		if(do_step)
		{
#ifdef BL_MP_COMB
			blMPGroupSync();
#else
			startTimer(&bl_time.comb);
			NestStep_combine(NULL,NULL);
			stopTimer(&bl_time.comb);
#endif
		}
	}
	if(do_step)
		NestStep_end_step();
	stopTimer(&bl_time.total);
}	

static BL_DTYPE (*globalMapTwoFunc)(BL_DTYPE,BL_DTYPE) = NULL;

static void localMapTwoGlobalFunc(BL_DTYPE *x1,BL_DTYPE *x2, BL_DTYPE *res, int n)
{
	int i;
	for(i=0;i<n;i++)
		res[i] = globalMapTwoFunc(x1[i], x2[i]);
}

void BL_TN(DistMapTwo)(BL_DTYPE (*map_func)(BL_DTYPE,BL_DTYPE), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* res, int N, enum bl_do_step do_step)
{
	globalMapTwoFunc = map_func;
	BL_TN(DistMapTwoLocalFunc)(&localMapTwoGlobalFunc, x1, x2, res, N, do_step);
}

void BL_TN(DistMapThreeLocalFunc)(void (*map_func)(BL_DTYPE*,BL_DTYPE*,BL_DTYPE*,BL_DTYPE*,int), BlockDistArray* x1,BlockDistArray* x2, BlockDistArray* x3, BlockDistArray* res, int N, enum bl_do_step do_step)
{
	startTimer(&bl_time.total);

	// init Barr handlers
	BArr_Handler baX1;
	init_BArr_Get(&baX1, x1, BL_CHSIZE);
	BArr_Handler baX2;
	init_BArr_Get(&baX2, x2, BL_CHSIZE);
	BArr_Handler baX3;
	init_BArr_Get(&baX3, x3, BL_CHSIZE);

	BArr_Handler baRes;
	init_BArr_Put(&baRes, res, BL_CHSIZE);
	
	if(do_step)
		NestStep_step();
    {	
		while(get_sw_BArr(&baX1)) // get chunks
		{
			get_sw_BArr(&baX2);
			get_sw_BArr(&baX3);
			
			startTimer(&bl_time.calc);
			map_func(baX1.current, baX2.current, baX3.current, baRes.current, baX1.currentSize);
			stopTimer(&bl_time.calc);
			
			put_sw_BArr(&baRes); // put res back
		}

		wait_put_BArr_done(&baRes); // wait for last put
		
		if(do_step)
		{
#ifdef BL_MP_COMB
			blMPGroupSync();
#else
			startTimer(&bl_time.comb);
			NestStep_combine(NULL,NULL);
			stopTimer(&bl_time.comb);
#endif
		}
	}
	if(do_step)
		NestStep_end_step();
	stopTimer(&bl_time.total);
}


static BL_DTYPE (*globalMapThreeFunc)(BL_DTYPE,BL_DTYPE,BL_DTYPE) = NULL;

static void localMapThreeGlobalFunc(BL_DTYPE *x1,BL_DTYPE *x2, BL_DTYPE *x3, BL_DTYPE *res, int n)
{
	int i;
	for(i=0;i<n;i++)
		res[i] = globalMapThreeFunc(x1[i], x2[i], x3[i]);
}


void BL_TN(DistMapThree)(BL_DTYPE (*map_func)(BL_DTYPE,BL_DTYPE,BL_DTYPE), BlockDistArray* x1, BlockDistArray* x2, BlockDistArray* x3, BlockDistArray* res, int N, enum bl_do_step do_step)
{
	globalMapThreeFunc = map_func;
	BL_TN(DistMapThreeLocalFunc)(&localMapThreeGlobalFunc, x1, x2, x3, res, N, do_step);
}
