#include "include/map_reduce.h"
#include "include/spu_timer.h"
#include "include/spu_utils.h"
#include "include/reduce.h"
#include "include/bl_f_d.h"

// this file is compiled both for floats and doubles
// Map and reduce combined


#ifndef BL_MP_COMB
extern RepArray* BL_TN(_bl_reduce_rep_arr);
#endif

BL_DTYPE BL_TN(DistMapOneAndReduceLocalFunc)(BL_DTYPE (*map_red_func)(BL_DTYPE*,int), BL_DTYPE (*red_func)(BL_DTYPE*,int),BlockDistArray* x, int N)
{
	startTimer(&bl_time.total);
	int i;
	BArr_Handler baX1;
	init_BArr_Get(&baX1, x, BL_CHSIZE);
	int rank = NestStep_get_rank();
	int groupSize = NestStep_get_size();
	BL_DTYPE red_res;
	BL_DTYPE _resVec[MAX_SPUS];
	for(i=0;i<groupSize;i++)
		_resVec[i]=0;
	BL_DTYPE res;
#ifndef BL_MP_COMB
	BL_TN(AllocReduceRepArray)();
#endif
	int redFirst=1;
	
	NestStep_step();
	{	
		while(get_sw_BArr(&baX1))
		{
			startTimer(&bl_time.calc);
			if(redFirst)
			{
				red_res = map_red_func(baX1.current, baX1.currentSize);
				redFirst = 0;
			}
			else
				red_res = BL_TN(ReduceTwo)(red_func, red_res, map_red_func(baX1.current, baX1.currentSize));
			stopTimer(&bl_time.calc);
		}
#ifdef BL_MP_COMB
		// sync local results
		blMPDistSync(_resVec, &red_res, BL_DTYPE_SIZE);
		startTimer(&bl_time.calc);
		// calculate global result
		res = red_func(_resVec, groupSize);
		stopTimer(&bl_time.calc);
#else
		_resVec[rank] = red_res;
		store_RArr(BL_TN(_bl_reduce_rep_arr), (Addr64)_resVec,0,MAX_SPUS-1);
		wait_RArr_complete(BL_TN(_bl_reduce_rep_arr),(Addr64)_resVec);
		NestStep_RepArray_attach(BL_TN(_bl_reduce_rep_arr), ADD, NULL, NO_PREFIX, NULL,NOT_UPDATED, NULL);
		startTimer(&bl_time.comb);
		NestStep_combine(NULL,NULL);
		stopTimer(&bl_time.comb);
	}
	NestStep_end_step();
	
	NestStep_step();
	{
		get_RArr(BL_TN(_bl_reduce_rep_arr), (Addr64)_resVec, 0, MAX_SPUS-1);
		startTimer(&bl_time.dma);
		wait_RArr_complete(BL_TN(_bl_reduce_rep_arr),(Addr64)_resVec);
		stopTimer(&bl_time.dma);
		startTimer(&bl_time.calc);
		res = red_func(_resVec, groupSize);
		stopTimer(&bl_time.calc);
		startTimer(&bl_time.comb);
		NestStep_combine(NULL,NULL);
		stopTimer(&bl_time.comb);
#endif
	}
	NestStep_end_step();
	
	stopTimer(&bl_time.total);
	return res;
}


// The code below is just a copy of the one above but with more argument vectors. Could be rewritten as 
// one set of functions with some more macro hacks or with pure c-kod and some wrapers for 
// stack-allocation of buffers.

BL_DTYPE BL_TN(DistMapTwoAndReduceLocalFunc)(BL_DTYPE (*map_red_func)(BL_DTYPE*,BL_DTYPE*,int), BL_DTYPE (*red_func)(BL_DTYPE*,int),BlockDistArray* x1,BlockDistArray* x2, int N)
{
	startTimer(&bl_time.total);	
	int i;
	BArr_Handler baX1;
	init_BArr_Get(&baX1, x1, BL_CHSIZE);
	BArr_Handler baX2;
	init_BArr_Get(&baX2, x2, BL_CHSIZE);

	int rank = NestStep_get_rank();
	int groupSize = NestStep_get_size();
	BL_DTYPE red_res;	
	BL_DTYPE _resVec[MAX_SPUS];
	for(i=0;i<groupSize;i++)
		_resVec[i]=0;
#ifndef BL_MP_COMB
	BL_TN(AllocReduceRepArray)();
#endif
	int sumFirst=1;
	BL_DTYPE res;	
	NestStep_step();
    {	
		while(get_sw_BArr(&baX1))
		{
			get_sw_BArr(&baX2);
			
			startTimer(&bl_time.calc);
			if(sumFirst)
			{
				red_res = map_red_func(baX1.current,baX2.current,baX1.currentSize);
				sumFirst = 0;
			}
			else
				red_res = BL_TN(ReduceTwo)(red_func, red_res, map_red_func(baX1.current,baX2.current,baX1.currentSize));
			stopTimer(&bl_time.calc);		
		}
#ifdef BL_MP_COMB
		blMPDistSync(_resVec, &red_res, BL_DTYPE_SIZE);
		startTimer(&bl_time.calc);
		res = red_func(_resVec, groupSize);
		stopTimer(&bl_time.calc);
#else
		_resVec[rank] = red_res;
		store_RArr(BL_TN(_bl_reduce_rep_arr), (Addr64)_resVec,0,MAX_SPUS-1);
		wait_RArr_complete(BL_TN(_bl_reduce_rep_arr),(Addr64)_resVec);
		NestStep_RepArray_attach(BL_TN(_bl_reduce_rep_arr), ADD, NULL, NO_PREFIX, NULL,NOT_UPDATED, NULL);
		startTimer(&bl_time.comb);
		NestStep_combine(NULL,NULL);
		stopTimer(&bl_time.comb);
	}	
	NestStep_end_step();

	NestStep_step();
	{
		get_RArr(BL_TN(_bl_reduce_rep_arr), (Addr64)_resVec, 0, MAX_SPUS-1);
		startTimer(&bl_time.dma);
		wait_RArr_complete(BL_TN(_bl_reduce_rep_arr),(Addr64)_resVec);
		stopTimer(&bl_time.dma);
		startTimer(&bl_time.calc);
		res = red_func(_resVec, groupSize);
		stopTimer(&bl_time.calc);
#endif
	}
	NestStep_end_step();

	stopTimer(&bl_time.total);
	return res;
}



BL_DTYPE BL_TN(DistMapThreeAndReduceLocalFunc)(BL_DTYPE (*map_red_func)(BL_DTYPE*,BL_DTYPE*,BL_DTYPE*,int), BL_DTYPE (*red_func)(BL_DTYPE*,int),BlockDistArray* x1,BlockDistArray* x2,BlockDistArray* x3, int N)
{
	startTimer(&bl_time.total);
	int i;
	BArr_Handler baX1;
	init_BArr_Get(&baX1, x1, BL_CHSIZE);
	BArr_Handler baX2;
	init_BArr_Get(&baX2, x2, BL_CHSIZE);
	BArr_Handler baX3;
	init_BArr_Get(&baX3, x3, BL_CHSIZE);

	int rank = NestStep_get_rank();
	int groupSize = NestStep_get_size();
	BL_DTYPE red_res;
	BL_DTYPE _resVec[MAX_SPUS];
	for(i=0;i<groupSize;i++)
		_resVec[i]=0;
	BL_DTYPE res;
#ifndef BL_MP_COMB
	BL_TN(AllocReduceRepArray)();
#endif
	int sumFirst=1;
	
	NestStep_step();
    {	
		while(get_sw_BArr(&baX1))
		{
			get_sw_BArr(&baX2);
			get_sw_BArr(&baX3);
			
			startTimer(&bl_time.calc);
			if(sumFirst)
			{
				red_res = map_red_func(baX1.current, baX2.current, baX3.current, baX1.currentSize);
				sumFirst = 0;
			}
			else
				red_res = BL_TN(ReduceTwo)(red_func, red_res, map_red_func(baX1.current,baX2.current, baX3.current, baX1.currentSize));
			stopTimer(&bl_time.calc);
			
		}
#ifdef BL_MP_COMB
		blMPDistSync(_resVec, &red_res, BL_DTYPE_SIZE);
		startTimer(&bl_time.calc);
		res = red_func(_resVec, groupSize);
		stopTimer(&bl_time.calc);
#else
		_resVec[rank] = red_res;
		store_RArr(BL_TN(_bl_reduce_rep_arr), (Addr64)_resVec,0,MAX_SPUS-1);
		wait_RArr_complete(BL_TN(_bl_reduce_rep_arr),(Addr64)_resVec);
		NestStep_RepArray_attach(BL_TN(_bl_reduce_rep_arr), ADD, NULL, NO_PREFIX, NULL,NOT_UPDATED, NULL);
		startTimer(&bl_time.comb);
		NestStep_combine(NULL,NULL);
		stopTimer(&bl_time.comb);
	}	
	NestStep_end_step();
	

	NestStep_step();
	{
		get_RArr(BL_TN(_bl_reduce_rep_arr), (Addr64)_resVec, 0, MAX_SPUS-1);
		startTimer(&bl_time.dma);
		wait_RArr_complete(BL_TN(_bl_reduce_rep_arr),(Addr64)_resVec);
		stopTimer(&bl_time.dma);
		startTimer(&bl_time.calc);
		res = red_func(_resVec, groupSize);
		stopTimer(&bl_time.calc);
#endif
	}
	NestStep_end_step();
	stopTimer(&bl_time.total);
	return res;
}

