#include "include/bl_mp.h"
#include "include/block_lib.h"


static volatile Ls_pointers ls[MAX_SPUS] __attribute__((aligned(16)));
static int lsInit = 0;
static volatile msgPointer msgData[MAX_SPUS] __attribute__((aligned(16)));
static volatile Dma_tag* msgSourceTags[MAX_SPUS] __attribute__((aligned(16)));
static msgSizes[MAX_SPUS];

// Get the other SPU:s addresses from the PPU. Will be cached
void initMp()
{
	if(!lsInit)
	{
		NestStep_get_all_ls(ls);
		lsInit = 1;
		int i;
		for(i=0;i<MAX_SPUS;i++)
		{
			msgSourceTags[i] = NULL;
			msgSizes[i] = 0;
		}
	}
}

void blSendSignal1(volatile unsigned int msg, int target)
{
	initMp();
	if(target == NestStep_get_rank())
	{
		prs("can't signal my self!");
		return;
	}
	volatile vector unsigned int volatile signal  __attribute__ ((aligned (128))) = {0,0,0,0};
	signal = spu_promote(msg, 3);
	mfc_write_tag_mask(1<<NESTSTEP_DMA_TAG);
	spu_dsync();
	// send the signal, mfc_sndsig and mfc_put acts almost identical, 
	// but mfc_put works better with the somwahat broken sdk 2.1
	//mfc_sndsig(((volatile void *)((volatile void *)&signal) + 12), 
	//			ls[target].sig1, NESTSTEP_DMA_TAG, 0, 0);
	mfc_put(((volatile void *)((volatile void *)&signal) + 12), 
				ls[target].sig1, 4, NESTSTEP_DMA_TAG, 0, 0);
	mfc_read_tag_status_all();
}

unsigned int blRecvSignal1()
{
	initMp();
	return spu_read_signal1();
}

void blSendSignal1Bit(int bit, int target)
{
	blSendSignal1(1<<bit, target);
}

// receive signals and or them into bitBuf, wait for bit 'bit' to arive.
// won't all bits received that is'nt 'bit' will be keept in bitBuf
static unsigned int bitBuf=0;
void blWaitSignal1Bit(int bit)
{
	while(!(bitBuf & (1<<bit)))
	{
		bitBuf |= blRecvSignal1();
	}
	bitBuf ^= (1<<bit);
}

// many of the small dumb functions here should scraped

void blSignal(int target)
{
	blSendSignal1Bit(NestStep_get_rank(), target);
}

void blSignalAck(int target)
{
	blSendSignal1Bit(NestStep_get_rank()+ACK_OFFSET, target);
}

void blWaitSignal(int source)
{
	blWaitSignal1Bit(source);
}

void blWaitSignalAck(int source)
{
	blWaitSignal1Bit(source+ACK_OFFSET);
}


void blSendMsg(volatile void * data, int size, int target)
{
	msgData[target].data = data;
	msgData[target].size = size;
	blSignal(target);
}

int blRecvMsg(volatile void * data, int source)
{
	initMp();
	if(!msgSourceTags[source]) // create a (neststep dma-) tag if there is none
		msgSourceTags[source] = new_tag();
	blWaitSignal(source);
	volatile msgPointer ptr __attribute__((aligned(16)));
	// get data pointer and array size
	mfc_get(&ptr, ls[source].ls+((unsigned int)&msgData[NestStep_get_rank()]), sizeof(msgPointer), NESTSTEP_DMA_TAG, 0, 0);
	mfc_write_tag_mask(1<<NESTSTEP_DMA_TAG);
	mfc_read_tag_status_all();
	msgSizes[source] = ptr.size;
	if(ptr.size) // dont get zero sized arrays
	{
		int tag = get_available_mfc_chan(); // get a free dma tag
		set_tag(msgSourceTags[source], tag, (Addr64)NULL);
		// start geting the array
		mfc_get(data,ls[source].ls+((unsigned int)ptr.data), ptr.size, tag,0,0);
	}
	return ptr.size;
}

void blWaitMsgDMA(int source)
{
	// messages of size zero are neither get nor acked
	if(msgSizes[source])
	{
		// wait for dma to finish
		mfc_write_tag_mask(1 << find_tag(msgSourceTags[source],(Addr64)NULL));
		startTimer(&bl_time.dma);
		mfc_read_tag_status_all();
		stopTimer(&bl_time.dma);
		// free dma tag
		dataManager.tags[find_tag(msgSourceTags[source],(Addr64)NULL)] = 0;
		remove_tag(msgSourceTags[source],(Addr64)NULL);
		// ack dma done
		blSendMsgAck(source);
	}
}

// wait for msg ack from reciever, i.e. that the whole msg is recieved 
// (and we can overwrite the buffer).
void blWaitMsgAck(int source)
{
	// msgs of size zero are not acked (used for chain termination)
	if(msgData[source].size) 
		blWaitSignalAck(source);
}

// syncronisation functions using message passing

// group sync using blMPGroupSync(); only if blocklib is compiled with the inter-spu-sync
void blGroupSync()
{
#ifdef BL_MP_COMB
	blMPGroupSync();
#else
	startTimer(&bl_time.comb);
	NestStep_combine(NULL,NULL);
	stopTimer(&bl_time.comb);
#endif
}

// group syncronization
void blMPGroupSync()
{
	// naive implementation but quite fast
	startTimer(&bl_time.comb);
	unsigned int rank = NestStep_get_rank();
	unsigned int groupSize = NestStep_get_size();
	int i;
	if(rank == 0)
	{
		for(i=1;i<groupSize;i++)
			blWaitSignal1Bit(16+i);
		for(i=1;i<groupSize;i++)
			blSendSignal1Bit(16,i);
	}
	else
	{
		blSendSignal1Bit(16+rank,0);
		blWaitSignal1Bit(16);
	}
	stopTimer(&bl_time.comb);
}

// syncronize and spread localElement. data must be a vector of at least 
// the 16 byte multiple bigger than group size * elementSize.
// Each spu get one spot in the array.
void blMPDistSync(volatile void* data, void* localElement, unsigned int elementSize)
{
	startTimer(&bl_time.comb);
	startTimer(&bl_time.exec);
	unsigned int rank = NestStep_get_rank();
	unsigned int groupSize = NestStep_get_size();
	
	int i;
	for(i=0;i<elementSize;i++)
		*((char*)(data+rank*elementSize+i)) = *((char*)localElement+i);
	


	/*
	// this version doesn't trigger the dma hang-bug in sdk 2.1
	int dataSize = elementSize*groupSize;
	if(dataSize%16)
		dataSize+=16-dataSize%16;

	if(rank == 0)
	{
		for(i=1;i<groupSize;i++)
			blRecvMsg((data+i*elementSize), i);
		for(i=1;i<groupSize;i++)
			blWaitMsgDMA(i);
		for(i=1;i<groupSize;i++)
			blSendMsg(data, dataSize, i);
		for(i=1;i<groupSize;i++)
			blWaitMsgAck(i);
	}
	else
	{
		blSendMsg((data+rank*elementSize), elementSize, 0);
		blWaitMsgAck(0);
		blRecvMsg((data), 0);
		blWaitMsgDMA(0);
	} */

	for(i=0;i<groupSize;i++)
		if(i!=rank)
			blSendMsg((data+rank*elementSize), elementSize, i);
	for(i=0;i<groupSize;i++)
		if(i!=rank)
			blRecvMsg((data+i*elementSize), i);
	for(i=0;i<groupSize;i++)
		if(i!=rank)
			blWaitMsgDMA(i);
	for(i=0;i<groupSize;i++)
		if(i!=rank)	
			blWaitMsgAck(i);

	stopTimer(&bl_time.comb);
}

