#include "include/bl_f_d.h"
#include "include/block_lib.h"

// this file is compiled both for floats and doubles

RepArray* BL_TN(_bl_reduce_rep_arr) = NULL;

void BL_TN(AllocReduceRepArray)()
{
	if(!BL_TN(_bl_reduce_rep_arr))
	{
		BL_DTYPE *ptr;
		Name name;
		name.procedure=NAME_OFFSET+NAME_REDUCE;
		name.relative=0;
		BL_TN(_bl_reduce_rep_arr) = new_RepArray(name, BL_NSTYPE, MAX_SPUS, (Addr64)ptr);
	}
}


BL_DTYPE BL_TN(DistReduceLocalFunc)(BL_DTYPE (*red_func)(BL_DTYPE*, int), BlockDistArray* x, int N)
{
	startTimer(&bl_time.total);
	int i;
	BArr_Handler ba;
	init_BArr_Get(&ba, x, BL_CHSIZE);
	int rank = NestStep_get_rank();
	int groupSize = NestStep_get_size();
	BL_DTYPE res;
	BL_DTYPE _resVec[MAX_SPUS];
	for(i=0;i<groupSize;i++)
		_resVec[i]=0;
#ifndef BL_MP_COMB
	BL_TN(AllocReduceRepArray)();
#endif
	int redFirst=1;
	NestStep_step();
    {	
		// reduce processors chunk
		while(get_sw_BArr(&ba))
		{
			startTimer(&bl_time.calc);
			if(redFirst) // first chunk?
			{
				res = red_func(ba.current,ba.currentSize); // reduce local array
				redFirst = 0;
			}
			else
			{	// reduce chunk and reduce {res,chunk_res}
				res = BL_TN(ReduceTwo)(red_func, res, red_func(ba.current,ba.currentSize));
			}
			stopTimer(&bl_time.calc);
		}
#ifdef BL_MP_COMB
		blMPDistSync(_resVec, &res, BL_DTYPE_SIZE); // get all other processors local results
		startTimer(&bl_time.calc);
		res = red_func(_resVec, groupSize); // reduce to a global result
		stopTimer(&bl_time.calc);
#else
		// same as above using NestStep Replicated Array
		_resVec[rank] = res;
		store_RArr(BL_TN(_bl_reduce_rep_arr), (Addr64)_resVec,0,MAX_SPUS-1);
		wait_RArr_complete(BL_TN(_bl_reduce_rep_arr),(Addr64)_resVec);
		NestStep_RepArray_attach(BL_TN(_bl_reduce_rep_arr), ADD, NULL, NO_PREFIX, NULL,NOT_UPDATED, NULL);
		startTimer(&bl_time.comb);
		NestStep_combine(NULL,NULL);
		stopTimer(&bl_time.comb);
	}	
	NestStep_end_step();

	NestStep_step();
	{
		startTimer(&bl_time.dma);
		get_RArr(BL_TN(_bl_reduce_rep_arr), (Addr64)_resVec, 0, MAX_SPUS-1);
		stopTimer(&bl_time.dma);
		wait_RArr_complete(BL_TN(_bl_reduce_rep_arr),(Addr64)_resVec);
		startTimer(&bl_time.calc);
		res = red_func(_resVec, groupSize);
		stopTimer(&bl_time.calc);
#endif
	}
	NestStep_end_step();

	stopTimer(&bl_time.total);
	return res;
}


static BL_DTYPE (*globalReduceFunc)(BL_DTYPE, BL_DTYPE) = NULL;

static BL_DTYPE localReduceGlobalFunc(BL_DTYPE* x, int N)
{
	int i=0;
	BL_DTYPE tmp=x[0];
	for(i=1;i<N;i++)
		tmp = globalReduceFunc(tmp, x[i]);
	return tmp;
}

BL_DTYPE BL_TN(DistReduce)(BL_DTYPE (*red_func)(BL_DTYPE, BL_DTYPE), BlockDistArray* x, int N)
{
	globalReduceFunc = red_func;
	return BL_TN(DistReduceLocalFunc)(&localReduceGlobalFunc, x, N);
}

BL_DTYPE BL_TN(LocalReduce)(BL_DTYPE (*red_func)(BL_DTYPE, BL_DTYPE), BL_DTYPE* x, int N)
{
	int i=0;
	BL_DTYPE tmp=x[0];
	for(i=1;i<N;i++)
		tmp = red_func(tmp, x[i]);
	return tmp;
}
