#include "include/spu_timer.h"
#include <spu_mfcio.h>
#include <stdio.h>

spu_timers bl_time={{0,0},{0,0},{0,0},{0,0},{0,0}};

void startTimer(spu_timer* timer)
{
	if(timer->startTime!=0)
		printf("startTime not 0!\n");
	timer->startTime = spu_read_decrementer();
}

void stopTimer(spu_timer* timer)
{
	if(timer->startTime==0)
		printf("startTime 0!\n");
	long long tmp = timer->startTime - spu_read_decrementer();
	if(tmp < 0)
		tmp += 4294967296; // 2^32 for wraparound
	timer->time += tmp;
	timer->startTime = 0;
}

float getTime(spu_timer* timer)
{
	return ((float)timer->time)/TIMEBASE;
}

// calculates the average of all timmers in timer
combTime combTimers(spu_timers timers)
{
	combTime times __attribute__((aligned(16)));
	Name name;
	name.procedure=NAME_OFFSET+NAME_TIME;
	name.relative=0;
	RepArray* rAvgTimes = new_RepArray(name, DVAR, 6, (Addr64)(double*)&times.avg);
	name.relative++;
	RepArray* rMinTimes = new_RepArray(name, DVAR, 6, (Addr64)(double*)&times.min);
	name.relative++;
	RepArray* rMaxTimes = new_RepArray(name, DVAR, 6, (Addr64)(double*)&times.max);
	name.relative++;
	
	times.avg.total=getTime(&timers.total);
	times.avg.exec=getTime(&timers.exec);
	times.avg.dma=getTime(&timers.dma);
	times.avg.comb=getTime(&timers.comb);
	times.avg.calc=getTime(&timers.calc);
	times.avg.un=getTime(&timers.total) - getTime(&timers.comb) - getTime(&timers.dma) - getTime(&timers.calc);
	times.max = times.avg;
	times.min = times.avg;
	NestStep_step();
	{
		store_RArrD(rAvgTimes, (Addr64)(double*)&times.avg,0,5);
		wait_RArr_complete(rAvgTimes,(Addr64)(double*)&times.avg);
		NestStep_RepArray_attach(rAvgTimes, ADD, NULL, NO_PREFIX, NULL,NOT_UPDATED, NULL);

		store_RArrD(rMinTimes, (Addr64)(double*)&times.min,0,5);
		wait_RArr_complete(rMinTimes,(Addr64)(double*)&times.min);
		NestStep_RepArray_attach(rMinTimes, COMBINE_MIN, NULL, NO_PREFIX, NULL,NOT_UPDATED, NULL);

		store_RArrD(rMaxTimes, (Addr64)(double*)&times.max,0,5);
		wait_RArr_complete(rMaxTimes,(Addr64)(double*)&times.max);
		NestStep_RepArray_attach(rMaxTimes, COMBINE_MAX, NULL, NO_PREFIX, NULL,NOT_UPDATED, NULL);
		
		NestStep_combine(NULL,NULL);
	}
	NestStep_end_step();
	NestStep_step();
	{
		get_RArrD(rAvgTimes, (Addr64)(double*)&times.avg, 0, 5);
		wait_RArr_complete(rAvgTimes,(Addr64)(double*)&times.avg);	
		
		get_RArrD(rMaxTimes, (Addr64)(double*)&times.max, 0, 5);
		wait_RArr_complete(rMaxTimes,(Addr64)(double*)&times.max);

		get_RArrD(rMinTimes, (Addr64)(double*)&times.min, 0, 5);
		wait_RArr_complete(rMinTimes,(Addr64)(double*)&times.min);
		
		NestStep_combine(NULL,NULL);
	}
	NestStep_end_step();
	int size = NestStep_get_size();
	times.avg.total/=size;
	times.avg.exec/=size;
	times.avg.dma/=size;
	times.avg.comb/=size;
	times.avg.calc/=size;
	times.avg.un/=size;
	
	free_RepArray(rAvgTimes);
	free_RepArray(rMaxTimes);
	free_RepArray(rMinTimes);
	NestStep_step();
	{
		NestStep_combine(NULL,NULL);
	}
	NestStep_end_step();
	return times;
}


void printTimersDetailed(spu_timers timers)
{
	combTime times = combTimers(timers);
	if(NestStep_get_rank()==0)
	{
		printf("Tot programtime:        %f\t %f\t %f\n", times.avg.exec, times.min.exec, times.max.exec);
		printf("Tot executiontime:      %f\t %f\t %f\n", times.avg.total, times.min.total, times.max.total);
		printf("Time spent calculating: %f\t %f\t %f\n", times.avg.calc, times.min.calc, times.max.calc);
		printf("Time spent waiting:     %f\t %f\t %f\n", times.avg.dma, times.min.dma, times.max.dma);
		printf("Time spent in combine:  %f\t %f\t %f\n", times.avg.comb, times.min.comb, times.max.comb);
		printf("Time unaccounted:       %f\t %f\t %f\n", times.avg.un, times.min.un, times.max.un);
		printf("\n");
	}
}

void printTimersDetailed_flops(spu_timers timers, long long n)
{
	combTime times = combTimers(timers);
	if(NestStep_get_rank()==0)
	{
		printf("Tot programtime:        %f\t %f\t %f\n", times.avg.exec, times.min.exec, times.max.exec);
		printf("Tot executiontime:      %f\t %f\t %f\t %fGF\n", times.avg.total, times.min.total, times.max.total, n/times.avg.total/(1000*1000*1000));
		printf("Time spent calculating: %f\t %f\t %f\t %fGF\n", times.avg.calc, times.min.calc, times.max.calc, n/times.avg.calc/(1000*1000*1000));
		printf("Time spent waiting:     %f\t %f\t %f\n", times.avg.dma, times.min.dma, times.max.dma);
		printf("Time spent in combine:  %f\t %f\t %f\n", times.avg.comb, times.min.comb, times.max.comb);
		printf("Time unaccounted:       %f\t %f\t %f\n", times.avg.un, times.min.un, times.max.un);
		printf("\n");
	}
}

void printTimers(spu_timers timers)
{
	printf("Tot programtime:        %f\n", getTime(&timers.exec));
	printf("Tot executiontime:      %f\n", getTime(&timers.total));
	printf("Time spent calculating: %f\n", getTime(&timers.calc));
	printf("Time spent waiting:     %f\n", getTime(&timers.dma));
	printf("Time spent in combine:  %f\n", getTime(&timers.comb));
	printf("Time unaccounted:       %f\n", getTime(&timers.total) - getTime(&timers.comb) - getTime(&timers.dma) - getTime(&timers.calc));
}

void printTimers_flops(spu_timers timers, long long n)
{
	printf("Tot programtime:        %f\n", getTime(&timers.exec));
	printf("Tot executiontime:      %f\t%f GF\n", getTime(&timers.total), n/getTime(&timers.total)/(1000*1000*1000));
	printf("Time spent calculating: %f\t%f GF\n", getTime(&timers.calc),n/getTime(&timers.calc)/(1000*1000*1000));
	printf("Time spent waiting:     %f\n", getTime(&timers.dma));
	printf("Time spent in combine:  %f\n", getTime(&timers.comb));
	printf("Time unaccounted:       %f\n", getTime(&timers.total) - getTime(&timers.comb) - getTime(&timers.dma) - getTime(&timers.calc));
}

