#include "include/spu_utils.h"

void* fix_align(unsigned int align, void* ptr)
{
	unsigned p = (unsigned int)ptr;
	if(p%align)
		p += align-(p%align);
	return (void*)p;
}

void get_BArrF_dA(BlockDistArray* _barr, float * lsaddr, int lbound, int ubound)
{
	int size = ubound - lbound +1;
	if(size%4 && size > 1) // do we have the not-a-factor-of-16b problem?
	{
		int i;
		int rest = size % 4;
		ubound -= rest; // fix the ubound for the big dma-get
		float * lsBase = &lsaddr[ubound - lbound + 1];
		for(i=0;i<rest; i++) 
		{
			get_BArrF(_barr, (Addr64)(&lsBase[i]), ubound+1+i, ubound+i+1);
			wait_BArr_complete(_barr,(Addr64)&lsBase[i]); // wait on every so we dont consume more dma-slots than needed
		}
	}
	if(ubound - lbound >= 0)
		get_BArrF(_barr, (Addr64)lsaddr, lbound, ubound);
}

void get_BArrD_dA(BlockDistArray* _barr, double * lsaddr, int lbound, int ubound)
{
	int size = ubound - lbound +1;
	if(size%2 && size > 1) // do we have the not-a-factor-of-16b problem?
	{
		get_BArrD(_barr, (Addr64)(&lsaddr[ubound-lbound]), ubound, ubound);
		wait_BArr_complete(_barr,(Addr64)&lsaddr[ubound-lbound]); // wait on every so we dont consume more dma-slots than needed
		ubound--;
	}
	get_BArrD(_barr, (Addr64)lsaddr, lbound, ubound);
}


void store_BArrF_dA(BlockDistArray* _barr, float* lsaddr, int lbound, int ubound)
{
	int size = ubound - lbound +1;
	if(size%4 && size > 1) // do we have the not-a-factor-of-16b problem?
	{
		int i;
		int rest = size % 4;
		ubound -= rest; // fix the ubound for the big dma-get;
		float * lsBase = &lsaddr[ubound - lbound + 1];
		for(i=0;i<rest; i++)
		{
			store_BArrF(_barr, (Addr64)&lsBase[i], ubound+1+i, ubound+i+1);
			wait_BArr_complete(_barr,(Addr64)&lsBase[i]);
		}
	}
	if(ubound - lbound >= 0)
		store_BArrF(_barr, (Addr64)lsaddr, lbound, ubound);
	
}

void store_BArrD_dA(BlockDistArray* _barr, double* lsaddr, int lbound, int ubound)
{
	int size = ubound - lbound +1;
	if(size%2 && size > 1) // do we have the not-a-factor-of-16b problem?
	{
		store_BArrD(_barr, (Addr64)&lsaddr[ubound-lbound], ubound, ubound);
		wait_BArr_complete(_barr,(Addr64)&lsaddr[ubound-lbound]);
		ubound--;
	}
	store_BArrD(_barr, (Addr64)lsaddr, lbound, ubound);
}

void init_BArrF_Get(BArrF_Handler *h, BlockDistArray* _a, int bs)
{
	h->a = _a;
	h->blockSize = bs;
	h->counter = 0;
	h->nextSize = 0;
	h->currentSize = 0;
	h->current = fix_align(BUF_ALIGN,h->buf1);
	h->next = fix_align(BUF_ALIGN,h->buf2);
	
	if(bs % 4)
		printf("Error: blocksize for BArrF Handler must be a factor of 4\n");

	h->part = getBDALocalSize(_a);
	h->start = getBDALowerBound(_a);
}

void init_BArrD_Get(BArrD_Handler *h, BlockDistArray* _a, int bs)
{
	h->a = _a;
	h->blockSize = bs;
	h->counter = 0;
	h->nextSize = 0;
	h->currentSize = 0;
	h->current = fix_align(BUF_ALIGN,h->buf1);
	h->next = fix_align(BUF_ALIGN,h->buf2);
	if(bs % 2)
		printf("Error: blocksize for BArrD Handler must be a factor of 2\n");

	h->part = getBDALocalSize(_a);
	h->start = getBDALowerBound(_a);
}

void init_BArrF_Put(BArrF_Handler *h, BlockDistArray* _a, int bs)
{
	h->part = getBDALocalSize(_a);
	h->start = getBDALowerBound(_a);
	h->a = _a;
	h->blockSize = bs;
	h->nextSize = 0;
	h->currentSize = (bs < h->part ? bs : h->part);
	h->currentStart = h->start;
	h->counter = h->currentSize;
	h->current = fix_align(BUF_ALIGN,h->buf1);
	h->next = fix_align(BUF_ALIGN,h->buf2);
	if(bs % 4)
		printf("Error: blocksize for BArrF Handler must be a factor of 4\n");
}

void init_BArrD_Put(BArrD_Handler *h, BlockDistArray* _a, int bs)
{
	h->part = getBDALocalSize(_a);
	h->start = getBDALowerBound(_a);
	h->a = _a;
	h->blockSize = bs;
	h->nextSize = 0;
	h->currentSize = (bs < h->part ? bs : h->part);
	h->currentStart = h->start;
	h->counter = h->currentSize;
	h->current = fix_align(BUF_ALIGN,h->buf1);
	h->next = fix_align(BUF_ALIGN,h->buf2);
	
	if(bs % 2)
		printf("Error: blocksize for BArrF Handler must be a factor of 2\n");
}

int get_sw_BArrF(BArrF_Handler *h)
{
	float *newNext = NULL;
	int newNextSize = 0;
	int newNextStart = 0;
	if(h->counter==0) // if this is first chunk, order first chunk to prime the pipeline
	{
		newNextSize = (h->blockSize < h->part ? h->blockSize : h->part);
		get_BArrF_dA(h->a, h->next, h->start, h->start + newNextSize -1);
		h->nextStart = h->start;
		h->nextSize = newNextSize;
		h->counter += newNextSize;
	}

	// wait for next
	startTimer(&bl_time.dma);
	wait_BArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);

	if(h->counter < h->part) // order next chunk
	{
		newNext = h->current;
		newNextSize = (h->blockSize < h->part - h->counter ? h->blockSize : h->part - h->counter);
		newNextStart = h->start + h->counter;
		get_BArrF_dA(h->a, newNext, h->start + h->counter, h->start + h->counter + newNextSize -1);
		h->counter += newNextSize;
	}
	else
		newNextSize = 0;

	h->current = h->next;
	h->currentSize = h->nextSize;
	h->currentStart = h->nextStart;
	
	h->next = newNext;
	h->nextSize = newNextSize;
	h->nextStart = newNextStart;

	return h->currentSize;
}

int get_sw_BArrD(BArrD_Handler *h)
{
	double *newNext = NULL;
	int newNextSize = 0;
	int newNextStart = 0;
	if(h->counter==0) // if this is first chunk, order first chunk to prime the pipeline
	{
		newNextSize = (h->blockSize < h->part ? h->blockSize : h->part);
		get_BArrD_dA(h->a, h->next, h->start, h->start + newNextSize -1);
		h->nextStart = h->start;
		h->nextSize = newNextSize;
		h->counter += newNextSize;
	}

	// wait for next
	startTimer(&bl_time.dma);
	wait_BArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);

	if(h->counter < h->part) // order next chunk
	{
		newNext = h->current;
		newNextSize = (h->blockSize < h->part - h->counter ? h->blockSize : h->part - h->counter);
		newNextStart = h->start + h->counter;
		get_BArrD_dA(h->a, newNext, h->start + h->counter, h->start + h->counter + newNextSize -1);
		h->counter += newNextSize;
	}
	else
		newNextSize = 0;

	h->current = h->next;
	h->currentSize = h->nextSize;
	h->currentStart = h->nextStart;
	
	h->next = newNext;
	h->nextSize = newNextSize;
	h->nextStart = newNextStart;

	return h->currentSize;
}

int put_sw_BArrF(BArrF_Handler *h)
{	
	// wait for last put to complete
	startTimer(&bl_time.dma);
	wait_BArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);

	float *tmp = h->next;
	int tmpSize;
	int tmpStart=0;

	if(h->counter < h->part)
	{
		tmpSize = (h->blockSize < h->part - h->counter ? h->blockSize : h->part - h->counter);
		tmpStart = h->start + h->counter;
		h->counter += tmpSize;
	}
	else
		tmpSize = 0;
	
	// swap
	h->next = h->current;
	h->nextSize = h->currentSize;
	h->nextStart = h->currentStart;
	h->current = tmp;
	h->currentSize = tmpSize;
	h->currentStart = tmpStart;
	
	// put new next
	store_BArrF_dA(h->a, h->next, h->nextStart, h->nextStart + h->nextSize-1);
	return h->currentSize;
}

int put_sw_BArrD(BArrD_Handler *h)
{	
	// wait for last put to complete
	startTimer(&bl_time.dma);
	wait_BArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);

	double *tmp = h->next;
	int tmpSize;
	int tmpStart=0;

	if(h->counter < h->part)
	{
		tmpSize = (h->blockSize < h->part - h->counter ? h->blockSize : h->part - h->counter);
		tmpStart = h->start + h->counter;
		h->counter += tmpSize;
	}
	else
		tmpSize = 0;
	
	// swap
	h->next = h->current;
	h->nextSize = h->currentSize;
	h->nextStart = h->currentStart;
	h->current = tmp;
	h->currentSize = tmpSize;
	h->currentStart = tmpStart;
	
	// put new next
	store_BArrD_dA(h->a, h->next, h->nextStart, h->nextStart + h->nextSize-1);
	return h->currentSize;
}

void wait_put_BArrF_done(BArrF_Handler *h)
{
	startTimer(&bl_time.dma);
	wait_BArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);
}

void wait_put_BArrD_done(BArrD_Handler *h)
{
	startTimer(&bl_time.dma);
	wait_BArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);
}

void printBArrF(BlockDistArray* x, int n, int nMax)
{
	if(n <= nMax)
	{
		int u,i;
		float buf[4096];
		int part = getBDALocalSize(x);
		int start = getBDALowerBound(x);

		for(u=0;u<NestStep_get_size();u++)
		{
			NestStep_step();
			{
				if(NestStep_get_rank() == u)
				{
					get_BArrF_dA(x,buf,start,start+part-1);
					wait_BArr_complete(x,(Addr64)buf);
					for(i=0;i<part;i++)
						printf("[%d]= %f\n", start+i, buf[i]);
				}		
				NestStep_combine(NULL,NULL);
			}
			NestStep_end_step();
		}
	}
}

void printBArrD(BlockDistArray* x, int n, int nMax)
{
	if(n <= nMax || nMax < 0)
	{
		int u,i;
		double buf[2048];
		int part = getBDALocalSize(x);
		int start = getBDALowerBound(x);
		BArrD_Handler baX;
		init_BArrD_Get(&baX, x, CHUNKSIZE_D);
		if(nMax <0)
			if(NestStep_get_rank() ==0)
				printf("RES: %d\n", n);
		for(u=0;u<NestStep_get_size();u++)
		{
			NestStep_step();
			{
				if(NestStep_get_rank() == u)
				{
					while(get_sw_BArrD(&baX))
					{
						if(nMax >0)
							for(i=0;i<baX.currentSize;i++)
								printf("[%d]= %f\n", baX.currentStart+i, baX.current[i]);
						else 
							for(i=0;i<baX.currentSize;i++)
								printf("RES: %.20e\n", baX.current[i]);
					}
				}
				NestStep_combine(NULL,NULL);
			}
			NestStep_end_step();
		}
	}
}

