#include "include/spu_utils.h"

// Private Array version of spu_utils.c

void get_PArrF_dA(PrivateArray* _parr, float * lsaddr, int lbound, int ubound)
{
	int size = ubound - lbound +1;
	if(size%4 && size > 1) // do we have the not-a-factor-of-16b problem?
	{
		int i;
		int rest = size % 4;
		ubound -= rest; // fix the ubound for the big dma-get
		float * lsBase = &lsaddr[ubound - lbound + 1];
		for(i=0;i<rest; i++) 
		{
			get_PArrF(_parr, (Addr64)(&lsBase[i]), ubound+1+i, ubound+i+1);
			wait_PArr_complete(_parr,(Addr64)&lsBase[i]); // wait on every so we dont consume more dma-slots than needed
		}
	}
	if(ubound - lbound >= 0) // anything left?
		get_PArrF(_parr, (Addr64)lsaddr, lbound, ubound);
}

void get_PArrD_dA(PrivateArray* _parr, double * lsaddr, int lbound, int ubound)
{
	int size = ubound - lbound +1;
	if(size%2 && size > 1) // do we have the not-a-factor-of-16b problem?
	{
		get_PArrD(_parr, (Addr64)(&lsaddr[ubound-lbound]), ubound, ubound);
		wait_PArr_complete(_parr,(Addr64)&lsaddr[ubound-lbound]); // wait on every so we dont consume more dma-slots than needed
		ubound--;
	}
	get_PArrD(_parr, (Addr64)lsaddr, lbound, ubound);
}


void store_PArrF_dA(PrivateArray* _parr, float* lsaddr, int lbound, int ubound)
{
	int size = ubound - lbound +1;
	if(size%4 && size > 1) // do we have the not-a-factor-of-16b problem?
	{
		int i;
		int rest = size % 4;
		ubound -= rest; // fix the ubound for the big dma-get;
		float * lsBase = &lsaddr[ubound - lbound + 1];
		for(i=0;i<rest; i++)
		{
			store_PArrF(_parr, (Addr64)&lsBase[i], ubound+1+i, ubound+i+1);
			wait_PArr_complete(_parr,(Addr64)&lsBase[i]);
		}
	}
	if(ubound - lbound >= 0)
		store_PArrF(_parr, (Addr64)lsaddr, lbound, ubound);
	
}

void store_PArrD_dA(PrivateArray* _parr, double* lsaddr, int lbound, int ubound)
{
	int size = ubound - lbound +1;
	if(size%2 && size > 1) // do we have the not-a-factor-of-16b problem?
	{
		store_PArrD(_parr, (Addr64)&lsaddr[ubound-lbound], ubound, ubound);
		wait_PArr_complete(_parr,(Addr64)&lsaddr[ubound-lbound]);
		ubound--;
	}
	store_PArrD(_parr, (Addr64)lsaddr, lbound, ubound);
}

void init_PArrF_Get(PArrF_Handler *h, PrivateArray* _a, int bs)
{
	h->a = _a;
	h->blockSize = bs;
	h->counter = 0;
	h->nextSize = 0;
	h->currentSize = 0;
	h->current = fix_align(BUF_ALIGN,h->buf1);
	h->next = fix_align(BUF_ALIGN,h->buf2);
	
	if(bs % 4)
		printf("Error: blocksize for PArrF Handler must be a factor of 4\n");

	h->size = getPASize(_a);
}


void init_PArrD_Get(PArrD_Handler *h, PrivateArray* _a, int bs)
{
	h->a = _a;
	h->blockSize = bs;
	h->counter = 0;
	h->nextSize = 0;
	h->currentSize = 0;
	h->current = fix_align(BUF_ALIGN,h->buf1);
	h->next = fix_align(BUF_ALIGN,h->buf2);
	if(bs % 2)
		printf("Error: blocksize for PArrD Handler must be a factor of 2\n");

	h->size = getPASize(_a);
}

void init_PArrF_Put(PArrF_Handler *h, PrivateArray* _a, int bs)
{
	h->size = getPASize(_a);
	h->a = _a;
	h->blockSize = bs;
	h->nextSize = 0;
	h->currentSize = (bs < h->size ? bs : h->size);
	h->currentStart = 0;
	h->counter = h->currentSize;
	h->current = fix_align(BUF_ALIGN,h->buf1);
	h->next = fix_align(BUF_ALIGN,h->buf2);
	
	if(bs % 4)
		printf("Error: blocksize for PArrF Handler must be a factor of 4\n");
}

void init_PArrD_Put(PArrD_Handler *h, PrivateArray* _a, int bs)
{
	h->size = getPASize(_a);
	h->a = _a;
	h->blockSize = bs;
	h->nextSize = 0;
	h->currentSize = (bs < h->size ? bs : h->size);
	h->currentStart = 0;
	h->counter = h->currentSize;
	h->current = fix_align(BUF_ALIGN,h->buf1);
	h->next = fix_align(BUF_ALIGN,h->buf2);
	
	if(bs % 2)
		printf("Error: blocksize for PArrD Handler must be a factor of 2\n");
}

int get_sw_PArrF(PArrF_Handler *h)
{
	float *newNext = NULL;
	int newNextSize = 0;
	int newNextStart = 0;
	if(h->counter==0) // first get, order and wait for first, order second
	{
		newNextSize = (h->blockSize < h->size ? h->blockSize : h->size);
		get_PArrF_dA(h->a, h->next, 0, newNextSize -1);
		h->nextStart = 0;
		h->nextSize = newNextSize;
		h->counter += newNextSize;
	}

	// wait for next
	startTimer(&bl_time.dma);
	wait_PArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);

	if(h->counter < h->size) // order next chunk
	{
		newNext = h->current;
		newNextSize = (h->blockSize < h->size - h->counter ? h->blockSize : h->size - h->counter);
		newNextStart = h->counter;
		get_PArrF_dA(h->a, newNext, h->counter, h->counter + newNextSize -1);
		h->counter += newNextSize;
	}
	else
		newNextSize = 0;

	h->current = h->next;
	h->currentSize = h->nextSize;
	h->currentStart = h->nextStart;
	
	h->next = newNext;
	h->nextSize = newNextSize;
	h->nextStart = newNextStart;

	return h->currentSize;
}

int get_sw_PArrD(PArrD_Handler *h)
{
	double *newNext = NULL;
	int newNextSize = 0;
	int newNextStart = 0;
	if(h->counter==0) // first get, order and wait for first, order second
	{
		newNextSize = (h->blockSize < h->size ? h->blockSize : h->size);
		get_PArrD_dA(h->a, h->next, 0, newNextSize -1);
		h->nextStart = 0;
		h->nextSize = newNextSize;
		h->counter += newNextSize;
	}

	// wait for next
	startTimer(&bl_time.dma);
	wait_PArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);

	if(h->counter < h->size) // order next chunk
	{
		newNext = h->current;
		newNextSize = (h->blockSize < h->size - h->counter ? h->blockSize : h->size - h->counter);
		newNextStart =  h->counter;
		get_PArrD_dA(h->a, newNext,  h->counter,  h->counter + newNextSize -1);
		h->counter += newNextSize;
	}
	else
		newNextSize = 0;

	h->current = h->next;
	h->currentSize = h->nextSize;
	h->currentStart = h->nextStart;
	
	h->next = newNext;
	h->nextSize = newNextSize;
	h->nextStart = newNextStart;

	return h->currentSize;
}

int put_sw_PArrF(PArrF_Handler *h)
{	
	// wait for last put to complete
	startTimer(&bl_time.dma);
	wait_PArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);

	float *tmp = h->next;
	int tmpSize;
	int tmpStart=0;

	if(h->counter < h->size)
	{
		tmpSize = (h->blockSize < h->size - h->counter ? h->blockSize : h->size - h->counter);
		tmpStart =  h->counter;
		h->counter += tmpSize;
	}
	else
		tmpSize = 0;
	
	// swap
	h->next = h->current;
	h->nextSize = h->currentSize;
	h->nextStart = h->currentStart;
	h->current = tmp;
	h->currentSize = tmpSize;
	h->currentStart = tmpStart;
	
	// put new next
	store_PArrF_dA(h->a, h->next, h->nextStart, h->nextStart + h->nextSize-1);
	return h->currentSize;
}

int put_sw_PArrD(PArrD_Handler *h)
{	
	// wait for last put to complete
	startTimer(&bl_time.dma);
	wait_PArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);

	double *tmp = h->next;
	int tmpSize;
	int tmpStart=0;

	if(h->counter < h->size)
	{
		tmpSize = (h->blockSize < h->size - h->counter ? h->blockSize : h->size - h->counter);
		tmpStart =  h->counter;
		h->counter += tmpSize;
	}
	else
		tmpSize = 0;
	
	// swap
	h->next = h->current;
	h->nextSize = h->currentSize;
	h->nextStart = h->currentStart;
	h->current = tmp;
	h->currentSize = tmpSize;
	h->currentStart = tmpStart;
	
	// put new next
	store_PArrD_dA(h->a, h->next, h->nextStart, h->nextStart + h->nextSize-1);
	return h->currentSize;
}

void wait_put_PArrF_done(PArrF_Handler *h)
{
	startTimer(&bl_time.dma);
	wait_PArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);
}

void wait_put_PArrD_done(PArrD_Handler *h)
{
	startTimer(&bl_time.dma);
	wait_PArr_complete(h->a,(Addr64)h->next);
	stopTimer(&bl_time.dma);
}

void printPArrF(PrivateArray* x, int n, int nMax)
{
	if(n <= nMax)
	{
		int i;
		float buf[1024];
		get_PArrF_dA(x,buf,0,n-1);
		wait_PArr_complete(x,(Addr64)buf);
		for(i=0;i<n;i++)
			printf("[%d]= %f\n", i, buf[i]);
	}
}

void printPArrD(PrivateArray* x, int n, int nMax)
{
	if(n <= nMax)
	{
		int i;
		double buf[2048];
		get_PArrD_dA(x,buf,0,n-1);
		wait_PArr_complete(x,(Addr64)buf);
		for(i=0;i<n;i++)
			printf("[%d]= %f\n", i, buf[i]);
	}
}

