/* **********************************************
 * BlockLib Test Program - Map
 * **********************************************/

 
#include <stdio.h>
#include <NestStep_spulib/neststep_spu.h>
#include "block_lib.h"
#include <spu_intrinsics.h>
// Initialization value for the counter
#define DEC_INIT_VAL 0XFFFFFFFF


#define MAX(a,b) (a>b ? a:b)

int N = 1024*1024*5;

int MY_RANK;

BlockDistArray* x;
BlockDistArray* y;
BlockDistArray* z;

float func(float left, float right)
{
	return MAX(left,right) *(left-right);
}

void func_local(float *x1, float *x2, float *res, int N)
{
	int i;
	for(i=0;i<N;i++)
		res[i] = MAX(x1[i],x2[i]) * (x1[i]- x2[i]);
}

float sqrt_func(float x)
{
	return sqrt(x);
}

float gen(int i)
{
	return (float)i;
}

float genMod(int i)
{
	return (float)(i%34);
}


DEF_GENERATE_FUNC_S(mGen, iF, BL_NONE, BL_NONE)

DEF_MAP_TWO_FUNC_S(map_func, t3, BL_NONE,
BL_SMAX(t1, op1, op2)
BL_SSUB(t2, op1, op2)
BL_SMUL(t3, t1,t2)
) 

float map_saxpy_alpha = 3;
DEF_MAP_TWO_FUNC_S(map_saxpy, t1, 
BL_SCONST(alpha, map_saxpy_alpha),
BL_SMUL_ADD(t1, op1, alpha, op2)
)

DEF_MAP_ONE_FUNC_S(sqrt_one, t1, BL_NONE,
BL_SSQRT(t1, op)
)

DEF_MAP_THREE_FUNC_S(map_matr, rq,
BL_SCONST(c11, 1.0f)
BL_SCONST(c12, 2.0f)
BL_SCONST(c13, 3.0f)
BL_SCONST(c14, 4.0f)
BL_SCONST(c21, 4.0f)
BL_SCONST(c22, 5.0f)
BL_SCONST(c23, 6.0f)
BL_SCONST(c24, 7.0f)
BL_SCONST(c31, 7.0f)
BL_SCONST(c32, 8.0f)
BL_SCONST(c33, 9.0f)
BL_SCONST(c34, 10.0f),

BL_SMUL_ADD(x11, op1, c11, c14)
BL_SMUL_ADD(x12, op1, c12, x11)
BL_SMUL_ADD(x1r, op1, c13, x12)
BL_SMUL(x1q, x1r, x1r)

BL_SMUL_ADD(x21, op2, c21, c24)
BL_SMUL_ADD(x22, op2, c22, x21)
BL_SMUL_ADD(x2r, op2, c23, x22)
BL_SMUL_ADD(x2q, x2r, x2r, x1q)

BL_SMUL_ADD(x31, op3, c31, c34)
BL_SMUL_ADD(x32, op3, c32, x31)
BL_SMUL_ADD(x3r, op3, c33, x32)
BL_SMUL_ADD(rq, x3r, x3r, x2q)
)


void calculate(){

	sDistGenerate(&gen, x, N,BL_STEP);
	sDistGenerate(&genMod, y, N, BL_STEP);
	sDistGenerate(&gen, z, N, BL_STEP);
	
	clearTimers(&bl_time);
	
	sDistGenerate(&gen, x, N,BL_STEP);

	printBArrF(x,N, 65);


	prs0("\nSimpel generator (1 ops/n)")
	prs0("opn: 1")
	printTimersDetailed_flops(bl_time, N);
	prs0("")

	blGroupSync();
	clearTimers(&bl_time);

		
	mGen(x, N,BL_STEP);

	printBArrF(x,N, 65);
	
	prs0("\nMacro generated generator (1 ops/n)")
	prs0("opn: 1")
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	
	blGroupSync();
	clearTimers(&bl_time);

	map_matr(x, y,z, z, N, BL_STEP);

	prs0("\nmap lenght in quadr after matrix transformation, macro generated (23 ops/n)")
	prs0("opn: 23")
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	
	blGroupSync();
	clearTimers(&bl_time);
	
	sDistMapTwo(&func, x, y, z, N, BL_STEP);

	//printBArrF(z,N);

	prs0("\nmap max(a,b)*(a-b), simple (4 ops/n)")
	prs0("opn: 4")
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	
	blGroupSync();
	clearTimers(&bl_time);

	sDistMapTwoLocalFunc(&func_local,x, y, z, N, BL_STEP);

	prs0("map max(a,b)*(a-b), local map function (4 ops/n)")
	prs0("opn: 4")
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	
	blGroupSync();
	clearTimers(&bl_time);
	
	map_func(x, y, z, N, BL_STEP);
	
	prs0("\nmap max(a,b)*(a-b), macro generated (4 ops/n)")
	prs0("opn: 4")
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	
	blGroupSync();
	clearTimers(&bl_time);
	
	sDistMapOne(&sqrt_func, x,z,N, BL_STEP);
		
	printBArrF(z,N, 65);
	
	prs0("\nmap one sqrt simple (1 sqrt/n)")
	prs0("opn: 1")
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	
	blGroupSync();
	clearTimers(&bl_time);
	
	sqrt_one(x,z,N,BL_STEP);
	
	prs0("\nmap one sqrt macro generated (1 sqrt/n)")
	prs0("opn: 1")
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	
	blGroupSync();
	clearTimers(&bl_time);

	map_saxpy_alpha = 3.5;
	map_saxpy(x,y,z,N, BL_STEP);

	printBArrF(z,N, 65);
	
	prs0("\nsaxpy macro generated (2 ops/n)")
	prs0("opn: 2")
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	clearTimers(&bl_time);
	
}

int main(unsigned long long speid, Addr64 argp, Addr64 envp)
{
    spu_write_decrementer(DEC_INIT_VAL);
    int i;
	
    NestStep_SPU_init(argp);
    MY_RANK = NestStep_get_rank();
    Name name;
    name.procedure = 1;
    name.relative = 0;

    x = new_BlockDistArray(name,FVAR,N);
	name.relative++;
	y = new_BlockDistArray(name,FVAR,N);
	name.relative++;
	z = new_BlockDistArray(name,FVAR,N);
	
	
    for(i=0;i<2;i++){
		if(MY_RANK==0)
			printf("\n\niteration %d:\n", i);
		calculate();
    }

	prd0(N)
	
    NestStep_SPU_finalize();

    return 0;
}

