/* **********************************************
 * BlockLib Test Program - Map-Reduce
 * **********************************************/
 
#include <stdio.h>
#include <NestStep_spulib/neststep_spu.h>
#include "block_lib.h"
#include <spu_intrinsics.h>
// Initialization value for the counter
#define DEC_INIT_VAL 0XFFFFFFFF

int N = 1024*1024*5;

int MY_RANK;

BlockDistArray* x;
BlockDistArray* y;
BlockDistArray* z;

float gen(int i)
{
	return (float)i;
}

DEF_MAP_THREE_AND_REDUCE_FUNC_S(matr_len, rq, t1,
BL_SCONST(c11, 1.0f)
BL_SCONST(c12, 2.0f)
BL_SCONST(c13, 3.0f)
BL_SCONST(c14, 4.0f)
BL_SCONST(c21, 4.0f)
BL_SCONST(c22, 5.0f)
BL_SCONST(c23, 6.0f)
BL_SCONST(c24, 7.0f)
BL_SCONST(c31, 7.0f)
BL_SCONST(c32, 8.0f)
BL_SCONST(c33, 9.0f)
BL_SCONST(c34, 10.0f),

BL_SMUL_ADD(x11, m_op1, c11, c14)
BL_SMUL_ADD(x12, m_op1, c12, x11)
BL_SMUL_ADD(x1r, m_op1, c13, x12)
BL_SMUL(x1q, x1r, x1r)

BL_SMUL_ADD(x21, m_op2, c21, c24)
BL_SMUL_ADD(x22, m_op2, c22, x21)
BL_SMUL_ADD(x2r, m_op2, c23, x22)
BL_SMUL_ADD(x2q, x2r, x2r, x1q)

BL_SMUL_ADD(x31, m_op3, c31, c34)
BL_SMUL_ADD(x32, m_op3, c32, x31)
BL_SMUL_ADD(x3r, m_op3, c33, x32)
BL_SMUL_ADD(rq, x3r, x3r, x2q),
BL_SMAX(t1,r_op1,r_op2)
)

DEF_MAP_TWO_FUNC_S(my_map, t1, BL_NONE,
BL_SMUL(t1, op1, op2)
)

DEF_REDUCE_FUNC_S(my_red,t1, BL_NONE,
BL_SADD(t1, op1, op2)
)

DEF_MAP_TWO_AND_REDUCE_FUNC_S(my_dot, mres, rres, BL_NONE, BL_SMUL(mres, m_op1, m_op2), BL_SADD(rres, r_op1, r_op2))

void calculate(){
	float res;
	sDistGenerate(&gen, x, N, BL_STEP);
	sDistGenerate(&gen, y, N, BL_STEP);
	sDistGenerate(&gen, z, N, BL_STEP);
	
	blGroupSync();
	clearTimers(&bl_time);
	
	res = my_dot(x,y,N);

	prs0("\nmap-reduce dot (2 ops/n)")
	prs0("opn: 2")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")	

	blGroupSync();
	clearTimers(&bl_time);

	my_map(x,y,z,N, BL_STEP);
	res= my_red(z,N);

	prs0("\nseperate map and reduce dot (2 ops/n)")
	prs0("opn: 2")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")	
	
	blGroupSync();
	clearTimers(&bl_time);

	res = matr_len(x,y,z,N);

	prs0("lenght of vector longest after matrix transformation (24 ops/n)\n")
	prs0("opn: 24")
	prf0(sqrt(res))
	printTimersDetailed_flops(bl_time, N);
	prs0("")	
}

int main(unsigned long long speid, Addr64 argp, Addr64 envp)
{
	spu_write_decrementer(DEC_INIT_VAL);
    int i;
	
    NestStep_SPU_init(argp);
    MY_RANK = NestStep_get_rank();
    Name name;
    name.procedure = 1;
    name.relative = 0;

    x = new_BlockDistArray(name,FVAR,N);
	name.relative++;
	y = new_BlockDistArray(name,FVAR,N);
	name.relative++;
	z = new_BlockDistArray(name,FVAR,N);
	
	
    for(i=0;i<2;i++){
		if(MY_RANK==0)
			printf("\n\niteration %d:\n", i);
		calculate();
    }

	prd0(N)
    NestStep_SPU_finalize();

    return 0;
}
