/* **********************************************
 * BlockLib Test Program - Overlapped Map
 * **********************************************/

 
#include <stdio.h>
#include <NestStep_spulib/neststep_spu.h>
#include "block_lib.h"
#include <spu_intrinsics.h>
#define DEC_INIT_VAL 0XFFFFFFFF


int N = 1024*1024*5;

int MY_RANK;

BlockDistArray* x;
BlockDistArray* xD;
BlockDistArray* y;
BlockDistArray* yD;


float gen(int i)
{
	return (float)i;
}

double gen_d(int i)
{
	return (double)i;
}

float genMod(int i)
{
	return (float)(i%34);
}


float overlapTest(float *x, int i)
{
	return x[-2]*0.1 + x[-1]*0.2 + x[0]*0.4 + x[1]*0.2 + x[2]*0.1;
}

void overlapLocal(float *x, float *res, int startN, int n)
{
	int i;
	for(i=0;i<n;i++)
		res[i] = x[i-2]*0.1 + x[i-1]*0.2 + x[i]*0.4 + x[i+1]*0.2 + x[i+2]*0.1;
}

double dOverlapTest(double *x, int i)
{
	return x[-2]*0.1 + x[-1]*0.2 + x[0]*0.4 + x[1]*0.2 + x[2]*0.1;
}

void dOverlapLocal(double *x, double *res, int startN, int n)
{
	int i;
	for(i=0;i<n;i++)
		res[i] = x[i-2]*0.1 + x[i-1]*0.2 + x[i]*0.4 + x[i+1]*0.2 + x[i+2]*0.1;
}

DEF_OVERLAPPED_MAP_ONE_FUNC_S(conv_s, vs5, 8, 
	BL_SCONST(p0, 0.4f)
	BL_SCONST(p1, 0.2f)
	BL_SCONST(p2, 0.1f),
	BL_SINDEXED(vm2, -2)
	BL_SINDEXED(vm1, -1)
	BL_SINDEXED(v, 0)
	BL_SINDEXED(vp1, 1)
	BL_SINDEXED(vp2, 2)
	BL_SMUL(vs1, vm2, p2)
	BL_SMUL_ADD(vs2, vm1, p1, vs1)
	BL_SMUL_ADD(vs3, v, p0, vs2)
	BL_SMUL_ADD(vs4, vp1, p1, vs3)
	BL_SMUL_ADD(vs5, vp2, p2, vs4)
)

DEF_OVERLAPPED_MAP_ONE_FUNC_D(conv_d, vs5, 8, 
	BL_DCONST(p0, 0.4)
	BL_DCONST(p1, 0.2)
	BL_DCONST(p2, 0.1),
	BL_DINDEXED(vm2, -2)
	BL_DINDEXED(vm1, -1)
	BL_DINDEXED(v, 0)
	BL_DINDEXED(vp1, 1)
	BL_DINDEXED(vp2, 2)
	BL_DMUL(vs1, vm2, p2)
	BL_DMUL_ADD(vs2, vm1, p1, vs1)
	BL_DMUL_ADD(vs3, v, p0, vs2)
	BL_DMUL_ADD(vs4, vp1, p1, vs3)
	BL_DMUL_ADD(vs5, vp2, p2, vs4)
)


DEF_GENERATE_FUNC_S(mGen, iF, BL_NONE, BL_NONE)
DEF_GENERATE_FUNC_D(mGen_d, iF, BL_NONE, BL_NONE)


void calculate(){
    int i, j, no_chunks;
    i = j = 0;

	mGen(x, N,BL_STEP);
	mGen(y, N,BL_STEP);
	mGen_d(xD, N, BL_STEP);
	mGen_d(yD, N, BL_STEP);

	
	if(MY_RANK == 0)
		printf("<<<<>>>>\n\n");

	clearTimers(&bl_time);
	
	sDistOverlappedMap(&overlapTest, x, y, N, 4, BL_ZERO);
//	sDistOverlappedMap(&overlapTest, x, y, N, 4, BL_CYCLIC);

	printBArrF(y,N, 65);
	
	prs0("\nSimple overlapped(single) (9 ops/n)")
	prs0("opn: 9")
	printTimersDetailed_flops(bl_time, N);
	prs0("")	

	blGroupSync();
	clearTimers(&bl_time);

	sDistOverlappedMapLocalFunc(&overlapLocal, x, y, N, 4, BL_ZERO);
//	sDistOverlappedMapLocalFunc(&overlapLocal, x, y, N, 4, BL_CYCLIC);

	printBArrF(y,N, 65);
	
	prs0("\nLocal overlapped (single) (9 ops/n)")
	prs0("opn: 9")
	printTimersDetailed_flops(bl_time, N);
	prs0("")	

	blGroupSync();
	clearTimers(&bl_time);

	conv_s(x,y,N,BL_ZERO);
//	conv_s(x,y,N,BL_CYCLIC);

	printBArrF(y,N, 65);
	
	prs0("\nMacro genterated overlapped (single) (9 ops/n)")
	prs0("opn: 9")
	printTimersDetailed_flops(bl_time, N);
	prs0("")	

	blGroupSync();
	clearTimers(&bl_time);
		
	dDistOverlappedMap(&dOverlapTest, xD, yD, N, 4, BL_ZERO);
//	dDistOverlappedMap(&dOverlapTest, xD, yD, N, 4, BL_CYCLIC);

	printBArrD(yD,N, 65);
	
	prs0("\nSimple overlapped(double) (9 ops/n)")
	prs0("opn: 9")
	printTimersDetailed_flops(bl_time, N);
	prs0("")	
	
	blGroupSync();
	clearTimers(&bl_time);
		
	dDistOverlappedMapLocalFunc(&dOverlapLocal, xD, yD, N, 4, BL_ZERO);
//	dDistOverlappedMapLocalFunc(&dOverlapLocal, xD, yD, N, 4, BL_CYCLIC);

	printBArrD(yD,N, 65);
	
	prs0("\nLocal overlapped (double) (9 ops/n)")
	prs0("opn: 9")
	printTimersDetailed_flops(bl_time, N);
	prs0("")	
	
	blGroupSync();   
	clearTimers(&bl_time);
	
	conv_d(xD,yD,N,BL_ZERO);
//	conv_d(xD,yD,N,BL_CYCLIC);

	printBArrD(yD,N, 65);
	
	prs0("\nMacro genterated overlapped (double) (9 ops/n)")
	prs0("opn: 9")
	printTimersDetailed_flops(bl_time, N);
	prs0("")	

	blGroupSync();
	clearTimers(&bl_time);
}

int main(unsigned long long speid, Addr64 argp, Addr64 envp)
{	
    spu_write_decrementer(DEC_INIT_VAL);
    int i;
	
    NestStep_SPU_init(argp);
    MY_RANK = NestStep_get_rank();
    Name name;
    name.procedure = 1;
    name.relative = 0;

    x = new_BlockDistArray(name,FVAR,N);
	name.relative++;
    xD = new_BlockDistArray(name,DVAR,N);
	name.relative++;
	y = new_BlockDistArray(name,FVAR,N);
	name.relative++;
	yD = new_BlockDistArray(name,DVAR,N);
	
    for(i=0;i<2;i++){
		if(MY_RANK==0)
			printf("\n\niteration %d:\n", i);
		calculate();
    }

    NestStep_SPU_finalize();

    return 0;
}

