/* **********************************************
 * BlockLib Test Program - Pipeline
 * **********************************************/

 
#include <stdio.h>
#include <NestStep_spulib/neststep_spu.h>
#include "block_lib.h"
#include <spu_intrinsics.h>

// Initialization value for the counter
#define DEC_INIT_VAL 0XFFFFFFFF



int rank;
int groupSize;
spu_timers timers={{0,0},{0,0},{0,0},{0,0},{0,0}};

void binPrint(unsigned int val)
{
	int u;
	char out[33];
	for(u=31;u>=0;u--)
		out[31-u]=(val&(1<<u) ? '1':'0');
	out[32]='\0';
	printf("(%d): %s\n", NestStep_get_rank(), out);	
}

DEF_GENERATE_FUNC_PRIVATE_S(gen, iF, BL_NONE, BL_NONE) 
DEF_GENERATE_FUNC_PRIVATE_S(genZero, zero, BL_SCONST(zero, 0.0f), BL_NONE) 

int N = 1024*1024*1;

#define CHSIZE 4096

DEF_GENERATE_FUNC_S(mGen, iF, BL_NONE, BL_NONE)

int main(unsigned long long speid, Addr64 argp, Addr64 envp)
{
	
    spu_write_decrementer(DEC_INIT_VAL);
    int i;
	
    NestStep_SPU_init(argp);
    rank = NestStep_get_rank();
	groupSize = NestStep_get_size();
	
    Name name;
    name.procedure = 1;
    name.relative = 0;
	PrivateArray* parr;

	if(rank == 0 || rank == groupSize -1)
	{
		parr = new_PrivateArray(name, FVAR, N, (Addr64)NULL);
	}
	name.relative++;
	
	BlockDistArray* x;
	x = new_BlockDistArray(name,FVAR,N*groupSize);
	mGen(x, N,BL_STEP);
	

	if(rank == 0)
		gen(parr, N, BL_NO_STEP);
	else if(rank == groupSize -1)
		genZero(parr, N, BL_NO_STEP);

	initMp();
	
	clearTimers(&bl_time);
	clearTimers(&timers);
	NestStep_step();
	{
		pipeF_Handler pipe; // pipe handler
		BArr_Handler baX; // block dist array handler
		for(i=0;i<1000;i++)
		{
			startTimer(&timers.total);
			
			// set the pipe up.
			if(rank == 0) 
				init_pipeF_first(&pipe, parr, rank+1, 4096);
			else if(rank != groupSize -1)
				init_pipeF_mid(&pipe, rank-1, rank+1, 4096);
			else
				init_pipeF_last(&pipe, NULL, rank-1, 4096);
			
			// init block dist array handler
			init_BArr_Get(&baX, x, 4096);

			while(step_PipeF(&pipe)) // loop trough all chunks
			{
				get_sw_BArr(&baX);
				// work on chunks here
			} 
			stopTimer(&timers.total);
		}
		NestStep_combine(NULL,NULL);
	}
	NestStep_end_step();
	prs("done");
	
	
	printTimersDetailed_flops(timers,N*1000ll*4);
	printTimersDetailed(bl_time);

	if(rank == 0 || rank == groupSize -1)
		free_PrivateArray(parr);
	free_BlockDistArray(x);
	
    NestStep_SPU_finalize();

    return 0;
} 

