/* **********************************************
 * BlockLib Test Program - Reduce
 * **********************************************/
 
#include <stdio.h>
#include <NestStep_spulib/neststep_spu.h>
#include "block_lib.h"
#include <spu_intrinsics.h>

// Initialization value for the counter
#define DEC_INIT_VAL 0XFFFFFFFF

int N = 1024*1024*5;

int MY_RANK;

BlockDistArray* x;

float add(float left, float right)
{
	return left+right;
}

float addReduce(float *x, int n)
{
	int i;
	float res=x[0];
	for(i=1;i<n;i++)
	{
		res+=x[i];
	}
	return res;
}

float addReduceVec(float *x, int n)
{
	int i;
	float res=0;
	int nVec=0;
	__align_hint(x,16,0);
	if(n>8)
	{	
		nVec = n/4;
		vector float vec_res __attribute__((aligned(16)));
		vec_res = spu_splats(0.0f);
		for(i=0;i<nVec;i++)
		{
			vec_res = spu_add(vec_res, ((vector float*)x)[i]);
		}
		res += spu_extract(vec_res, 0) + spu_extract(vec_res, 1) + spu_extract(vec_res, 2) + spu_extract(vec_res, 3);
	}
	for(i=nVec*4;i<n;i++)
	{
		res+=x[i];
	}
	return res;
}


float gen(int i)
{
	return (float)i;
}

DEF_REDUCE_FUNC_S(max, t1, BL_NONE,
BL_SMAX(t1, op1, op2)
)

DEF_REDUCE_FUNC_S(my_add_red, t1, BL_NONE,
BL_SADD(t1, op1, op2)
)

void calculate(){
    int i, j, no_chunks;
    i = j = 0;
	float res;

	sDistGenerate(&gen, x, N, BL_STEP);

	clearTimers(&bl_time);
 	res = sDistReduce(&add, x, N);
		
	prs0("\nSimple sum (1 ops/n)")
	prs0("opn: 1")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	
	blGroupSync();
	clearTimers(&bl_time);
		
	res = sDistReduceLocalFunc(&addReduce, x, N);
	
	prs0("\nLocal sum (1 ops/n)")
	prs0("opn: 1")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")	

	blGroupSync();
	clearTimers(&bl_time);
	
	res = sDistReduceLocalFunc(&addReduceVec, x, N);
	
	prs0("\nLocal vectorized sum (1 ops/n)")
	prs0("opn: 1")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")	

	blGroupSync();
	clearTimers(&bl_time);
	
	res = my_add_red(x, N);
	
	prs0("\nLocal macro generated sum (1 ops/n)")
	prs0("opn: 1")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")	

	blGroupSync();
	clearTimers(&bl_time);;

	res = max(x, N);
	
	prs0("\nLocal macro generated max (2 ops/n)")
	prs0("opn: 2")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")	
	clearTimers(&bl_time);;
}

int main(unsigned long long speid, Addr64 argp, Addr64 envp)
{
    spu_write_decrementer(DEC_INIT_VAL);
    int i;


	NestStep_SPU_init(argp);
    MY_RANK = NestStep_get_rank();
    Name name;
    name.procedure = 1;
    name.relative = 0;

    x = new_BlockDistArray(name,FVAR,N);
    for(i=0;i<2;i++){
		if(MY_RANK==0)
			printf("\n\niteration %d:\n", i);
		calculate();
    }

	prd0(N)
	
    NestStep_SPU_finalize();

    return 0;
}
