/* **********************************************
 * BlockLib Test Program - Reduce (double)
 * **********************************************/

#include <stdio.h>
#include <NestStep_spulib/neststep_spu.h>
#include "block_lib.h"
#include <spu_intrinsics.h>
#define DEC_INIT_VAL 0XFFFFFFFF

int N = 1024*1024*5;

int MY_RANK;

BlockDistArray* x;

double add(double left, double right)
{
	return left+right;
}

double addReduce(double *x, int n)
{
	int i;
	double res=x[0];
	for(i=1;i<n;i++)
	{
		res+=x[i];
	}
	return res;
}

double addReduceVec(double *x, int n)
{
	int i;
	double res=0;
	int nVec=0;
	if(n>4)
	{	
		nVec = n/2;
		vector double vec_res __attribute__((aligned(16)));
		vec_res = spu_splats(0.0);
		for(i=0;i<nVec;i++)
		{
			vec_res = spu_add(vec_res, ((vector double*)x)[i]);
		}
		res += spu_extract(vec_res, 0) + spu_extract(vec_res, 1);
	}
	for(i=nVec*2;i<n;i++)
	{
		res+=x[i];
	}
	return res;
}


double gen(int i)
{
	return (double)i;
}


DEF_REDUCE_FUNC_D(max, t1, BL_NONE,
BL_DMAX(t1, op1, op2)
)

DEF_REDUCE_FUNC_D(my_add_red, t1, BL_NONE,
BL_DADD(t1, op1, op2)
)

void calculate(){
	double res;
 	dDistGenerate(&gen, x, N, BL_STEP);

	clearTimers(&bl_time);
	res = dDistReduce(&add, x, N);
		
	prs0("\nSimple sum(1 ops/n)")
	prs0("opn: 1")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")
	
	blGroupSync();
	clearTimers(&bl_time);
	
	res = dDistReduceLocalFunc(&addReduce, x, N);
	
	prs0("\nLocal sum (1 ops/n)")
	prs0("opn: 1")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")

	blGroupSync();
	clearTimers(&bl_time);

	res = dDistReduceLocalFunc(&addReduceVec, x, N);
	
	prs0("\nLocal vectorized sum (1 ops/n)")
	prs0("opn: 1")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")	

	blGroupSync();
	clearTimers(&bl_time);
	
	res = my_add_red(x, N);
	
	prs0("\nLocal macro generated sum (1 ops/n)")
	prs0("opn: 1")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")	
	
	blGroupSync();
	clearTimers(&bl_time);
	
	res = max(x, N);
	
	prs0("\nLocal macro generated max (2 ops/n)")
	prs0("opn: 2")
	prf0(res)
	printTimersDetailed_flops(bl_time, N);
	prs0("")	
}

int main(unsigned long long speid, Addr64 argp, Addr64 envp)
{
    spu_write_decrementer(DEC_INIT_VAL);
    int i;

	NestStep_SPU_init(argp);
    MY_RANK = NestStep_get_rank();
    Name name;
    name.procedure = 1;
    name.relative = 0;

    x = new_BlockDistArray(name,DVAR,N);
    for(i=0;i<2;i++){
		if(MY_RANK==0)
			printf("\n\niteration %d:\n", i);
		calculate();
    }

	prd0(N)
	
    NestStep_SPU_finalize();

    return 0;
}
